from sensiml.datamanager.pipeline import PipelineStep


class QueryCall(PipelineStep):
    """The base class for a query call."""

    def __init__(self, name):
        super(QueryCall, self).__init__(name=name, step_type="QueryCall")
        self._query = None

    @property
    def query(self):
        return self._query

    @query.setter
    def query(self, query):
        self._query = query

    @property
    def outputs(self):
        return self._outputs

    @outputs.setter
    def outputs(self, outputs):
        self._outputs = outputs

    def _to_dict(self):
        query_dict = {}
        query_dict["name"] = self.query.name
        query_dict["type"] = "query"
        query_dict["outputs"] = getattr(self, "_outputs")
        return query_dict



import os
import json
import IPython
from collections import namedtuple
from operator import attrgetter
from pandas import DataFrame
from ipywidgets import widgets
from ipywidgets import (
    Layout,
    Button,
    VBox,
    HBox,
    Box,
    FloatText,
    Textarea,
    Dropdown,
    Label,
    Tab,
    IntSlider,
    Checkbox,
    Text,
    Button,
    SelectMultiple,
    HTML,
)
from IPython.display import display
from ipywidgets import IntText
from json import dumps as jdump
from sensiml.widgets.base_widget import BaseWidget
from sensiml.widgets.renderers import WidgetAttributeRenderer


category_item_layout = Layout(
    # display='flex',
    size=16,
    # border='solid 2px',
    justify_content="flex-start",
    # background_color= 'red',
    overflow="visible",
)


def clean_name(name):
    return "".join(e if e.isalnum() else "_" for e in name)


class DownloadWidget(BaseWidget):
    def __init__(self, dsk=None, level="Project", folder="knowledgepacks"):
        self._dsk = dsk
        if not os.path.exists(folder):
            os.makedirs(folder)
        self.setup(level=level)

    def setup(self, level):
        self.kb_description = {"parent": {}, "sub": {"Report": "Report"}}

        self.kb_dict = {"parent": [], "sub": []}
        self.level = level

    def select_platform(self, b):
        platform = self._dsk.platforms.get_platform_by_id(self._widget_platform.value)
        supported_source_drivers = platform.supported_source_drivers
        supported_source_keys = {x: x for x in list(supported_source_drivers.keys())}

        capture_configurations = {
            k: v.uuid
            for k, v in self._dsk.project.capture_configurations.build_capture_list().items()
        }

        supported_source_keys.update(capture_configurations)
        supported_source_keys["Custom"] = "Custom"

        default_source_driver = supported_source_drivers.get("Default", None)
        supported_outputs = platform.supported_outputs

        if default_source_driver is not None:
            # Remove the Default option from the list.
            supported_source_keys.pop("Default")

        self._widget_source.options = supported_source_keys
        if len(self._widget_source.options) > 0:
            if capture_configurations:
                self._widget_source.value = next(iter(capture_configurations.items()))[
                    1
                ]
            else:
                self._widget_source.value = "Custom"
        else:
            self._widget_source.options = {"Custom": "Custom"}
            self._widget_source.value = self._widget_source.options[0]
        self._widget_app_outputs.options = supported_outputs
        self._widget_app_outputs.value = supported_outputs
        self._change_rate("value")
        if len(self._widget_source_rate.options) > 0:
            self._widget_source_rate.value = self._widget_source_rate.options[0]
        else:
            self._widget_source_rate.options = [""]
            self._widget_source_rate.value = ""
        self._widget_target_os.options = platform.target_os_options
        self._widget_target_os.value = self._widget_target_os.options[0]

        if platform.can_build_binary:
            self._widget_download_type.options = ["Binary", "Library"]
        else:
            self._widget_download_type.options = ["Library"]
        self._widget_download_type.value = self._widget_download_type.options[0]

    def generate_description(self, b):
        if hasattr(self._dsk, "_auto_sense_ran"):
            if self._dsk._auto_sense_ran:
                self.renderer.render(
                    "KnowledgePack List Needs updating. Updating Now... Reselect and redownload."
                )
                self._refresh_models_list(None)
                self._dsk._auto_sense_ran = False
                return

        platform = self._dsk.platforms.get_platform_by_id(self._widget_platform.value)
        supported_outputs = platform.supported_outputs
        parent_name = None
        parent_uuid = None
        description = {}
        self.set_parent_model(None)
        if not self.kb_description["parent"]:
            self.renderer.render("No Parent Model Selected.")
            return
        for key in self.kb_description["parent"]:
            description = {
                clean_name(key): {
                    "uuid": self.kb_description["parent"][key],
                    "results": {},
                    "source": self._widget_source.value,
                }
            }
            parent_name = clean_name(key)
            parent_uuid = self.kb_description["parent"][key]
        for parent in self.kb_dict["parent"][0][1:]:
            description[parent_name]["results"].update(
                {format(parent.description.split("-")[0]): clean_name(parent.value)}
            )
        for key in self.kb_description["sub"]:
            if key != "Report":
                sub_description = {
                    "uuid": self.kb_description["sub"][key],
                    "parent": parent_name,
                    "segmenter_from": "parent",
                }
                description.update({clean_name(key): sub_description})
        board_name = self._dsk.platforms.get_platform_by_id(
            self._widget_platform.value
        ).board_name

        kp_uuid = parent_uuid
        kp_platform = self._widget_platform.value
        kp_debug = self._widget_debug.value
        kp_test_data = self._widget_test_data.value
        kp_download_type = self._widget_download_type.value

        debug_level = self._widget_debug_level.value
        profile = self._widget_profile.value
        profile_iterations = self._widget_profile_iterations.value

        sample_rate = (
            self._widget_source_rate.value
            if self._widget_source_rate.value in self._widget_source_rate.options
            else 100
        )
        output_options = [x.lower() for x in self._widget_app_outputs.value]

        execution_parameters = platform.execution_parameters

        if "ble" in supported_outputs and "ble" not in output_options:
            output_options.insert("ble")

        kp_application = execution_parameters.get("application", "Default")

        if kp_uuid is not None:
            kp = self._dsk.get_knowledgepack(kp_uuid)
        else:
            return None

        if kp.knowledgepack_description is not None:
            description = kp.knowledgepack_description
            description["Parent"]["source"] = self._widget_source.value

        config = {
            "target_platform": kp_platform,
            "test_data": kp_test_data,
            "debug": kp_debug,
            "application": kp_application,
            "sample_rate": sample_rate,
            "output_options": output_options,
            "kb_description": description,
            "debug_level": debug_level,
            "profile": profile,
            "profile_iterations": profile_iterations,
        }
        print(config)

        self.renderer.render("Generating Knowledge Pack")

        if kp_download_type == "Library":
            save_path, saved = kp.download_library(
                config=config, folder="knowledgepacks", renderer=self.renderer
            )
        if kp_download_type == "Binary":
            save_path, saved = kp.download_binary(
                config=config, folder="knowledgepacks", renderer=self.renderer
            )

    def set_parent_model(self, b):

        if self._widget_parent_select.value is None:
            return

        kp = self._dsk.get_knowledgepack(self._widget_parent_select.value)
        kp_item = []
        kp_item.append(Label(value=kp.name))

        for key, value in kp.class_map.items():
            kp_item.append(
                Dropdown(
                    options=self.kb_description["sub"],
                    description="{} - {}".format(key, value),
                )
            )

        self.kb_description["parent"] = {kp.name: kp.uuid}
        self.kb_dict["parent"] = [kp_item]
        self.update_models()

    def get_kp_dict(self):
        if self.level.lower() == "project":
            kps = self._dsk.project.list_knowledgepacks()
        elif self.level.lower() == "pipeline":
            kps = self._dsk.pipeline.list_knowledgepacks()
        else:
            kps = self._dsk.list_knowledgepacks()

        if isinstance(kps, DataFrame) and len(kps):
            kps = sorted(
                [
                    (name, value)
                    for name, value in kps[["Name", "kp_uuid"]].values
                    if name
                ],
                key=lambda s: s[0].lower(),
            )

            return kps

        return [("", None)]

    @staticmethod
    def flatten(l):
        return [item for sublist in l for item in sublist]

    def get_model_list(self):
        return (
            [Label(value="Parent Model")]
            + self.flatten(self.kb_dict["parent"])
            + [Label(value="Sub Model")]
            + self.flatten(self.kb_dict["sub"])
        )

    def get_feature_file_list(self):
        ff = self._dsk.list_featurefiles(silent=True)
        if ff is not None:
            return list(ff["Name"].values)
        else:
            return []

    def get_platform_names(self):
        named_platform = namedtuple("Platform", ["name", "version", "id"])
        platform_list = []
        for platform in self._dsk.platforms.platform_list:
            platform_list.append(
                named_platform(
                    platform.platform, platform.platform_version, platform.id
                )
            )
        platform_list = sorted(platform_list, key=attrgetter("name", "version"))
        ret_dict = {}
        for platform in platform_list:
            ret_dict["{} {}".format(platform.name, platform.version)] = platform.id

        return ret_dict

    def _refresh(self):
        if self._dsk is None:
            return
        self._widget_platform.options = self.get_platform_names()
        # self._widget_platform.value = 10
        self._widget_platform.observe(self.select_platform)
        self.select_platform(None)
        self._widget_test_data.options = [None] + self.get_feature_file_list()
        self._widget_parent_select.options = self.get_kp_dict()
        self._widget_parent_select.value = self._widget_parent_select.options[0][1]
        self._widget_class_map.options = self._get_class_map()

    def _clear(self):
        self._widget_parent_select.options = [""]
        self._widget_parent_select.value = ""
        self._widget_class_map.options = []

    def _update_class_map(self, *args):
        self._widget_class_map.options = self._get_class_map()

    def _get_class_map(self):

        if self._widget_parent_select.value:
            class_map = self._dsk.get_knowledgepack(
                self._widget_parent_select.value
            ).class_map
            return sorted(
                ["{} - {}".format(key, value) for key, value in class_map.items()]
            )
        else:
            return ""

    def _refresh_models_list(self, b):
        if self._dsk:
            if self._dsk.pipeline:
                self._widget_parent_select.options = self.get_kp_dict()
                self._widget_parent_select.value = self._widget_parent_select.options[
                    0
                ][1]
                self._widget_class_map.options = self._get_class_map()

    def update_models(self):
        if self._dsk is None:
            return

        if self.kb_dict["parent"]:
            for output in self.kb_dict["parent"][0][1:]:
                output.options = [k for k, v in self.kb_description["sub"].items()]

    def _change_rate(self, change):
        platform = self._dsk.platforms.get_platform_by_id(self._widget_platform.value)
        supported_source_drivers = platform.supported_source_drivers

        if self._widget_source.value == "Custom":
            self._widget_source_rate.options = [""]
            self._widget_source_rate.value = None
        else:
            self._widget_source_rate.options = supported_source_drivers.get(
                self._widget_source.value, [""]
            )

        if self._widget_source_rate.options[0] == "":
            self._widget_source_rate.layout.visibility = "hidden"
        else:
            self._widget_source_rate.layout.visibility = "visible"

    def create_widget(self):

        info_layout = Layout(width="35px", visibility="visible")

        description_style = {"description_width": "112px"}

        self._info_test_data = widgets.Button(
            icon="question",
            disabled=True,
            tooltip="""
                    Select a test file to upload. The knowledge pack will replace incoming sensor
                    data with this test data. Allows for validating model performance on the device 
                    with known results. Once the end of the test data is reached, the knowledge pack
                    will start from the start of the test data again.
                    """,
            layout=info_layout,
        )

        self._info_debug = widgets.Button(
            icon="question",
            disabled=True,
            tooltip="""
                    Enable Debug output over serial. 
                    """,
            layout=info_layout,
        )

        self._info_debug_level = widgets.Button(
            icon="question",
            disabled=True,
            tooltip="""
                    Debug Level 1: Classification 
                    Debug Level 2: Feature Vector 
                    Debug Level 3: Raw Data Fed into Feature Extractors
                    """,
            layout=info_layout,
        )
        self._info_profile = widgets.Button(
            icon="question",
            disabled=True,
            tooltip="""
                    Adds profiling to code which will output the MIPS and Average seconds for
                    each feature extractor and classifier.
                    """,
            layout=info_layout,
        )

        self._info_profile_iterations = widgets.Button(
            icon="question",
            disabled=True,
            tooltip="""
                    The number of iterations to perform when profiling each feature extrator and classifier.
                    """,
            layout=info_layout,
        )

        self._button_generate = Button(
            icon="download", tooltip="Generate and Download", layout=Layout(width="15%")
        )
        self._button_refresh = Button(
            icon="refresh", layout=Layout(width="15%"), tooltip="Refresh Model List"
        )
        self._widget_platform = Dropdown(description="HW Platform", options=[])
        self._widget_target_os = Dropdown(description="Target OS")
        self._widget_app_outputs = SelectMultiple(description="Output", options=[])
        self._widget_download_type = Dropdown(description="Format", options=[])
        self._widget_source = Dropdown(description="Data Source", options=[])
        self._widget_source_rate = Dropdown(description="Sample Rate", options=[])
        self._widget_debug = Dropdown(
            description="Debug", options=[False, True], style=description_style
        )
        self._widget_debug_level = Dropdown(
            description="Debug Level", options=[1, 2, 3], style=description_style
        )
        self._widget_profile = Dropdown(
            description="Profile", options=[False, True], style=description_style
        )
        self._widget_profile_iterations = IntText(
            description="Profile Iterations", default=1000, style=description_style
        )
        self._widget_test_data = Dropdown(
            description="Test Data", options=[None], style=description_style
        )
        self._widget_parent_select = Dropdown(
            description="Model Name", options=[], layout=Layout(width="85%")
        )
        self._widget_class_map = SelectMultiple(description="Class Map", options=[""])

        self._widget_render_space = HTML()

        self.renderer = WidgetAttributeRenderer(self._widget_render_space, "value")

        self.kb_items = VBox(
            [
                Tab(
                    [
                        VBox(
                            [
                                HBox(
                                    [
                                        self._widget_parent_select,
                                        self._button_refresh,
                                        self._button_generate,
                                    ]
                                ),
                                self._widget_class_map,
                                Label(
                                    value="Target Device Options",
                                    layout=category_item_layout,
                                ),
                                HBox(
                                    [
                                        VBox(
                                            [
                                                self._widget_platform,
                                                self._widget_target_os,
                                                self._widget_download_type,
                                                self._widget_source,
                                                self._widget_source_rate,
                                            ]
                                        ),
                                        self._widget_app_outputs,
                                    ]
                                ),
                            ]
                        ),
                        HBox(
                            [
                                VBox(
                                    [
                                        HBox(
                                            [
                                                self._widget_test_data,
                                                self._info_test_data,
                                            ],
                                            layout=Layout(align_self="flex-start"),
                                        ),
                                        HBox(
                                            [self._widget_debug, self._info_debug],
                                            layout=Layout(align_self="flex-start"),
                                        ),
                                        HBox(
                                            [
                                                self._widget_debug_level,
                                                self._info_debug_level,
                                            ],
                                            layout=Layout(align_self="flex-start"),
                                        ),
                                        HBox(
                                            [self._widget_profile, self._info_profile],
                                            layout=Layout(align_self="flex-start"),
                                        ),
                                        HBox(
                                            [
                                                self._widget_profile_iterations,
                                                self._info_profile_iterations,
                                            ],
                                            layout=Layout(align_self="flex-start"),
                                        ),
                                    ]
                                )
                            ]
                        ),
                    ]
                ),
                self._widget_render_space,
            ]
        )

        self.kb_items.children[0].set_title(0, "Main Settings")
        self.kb_items.children[0].set_title(1, "Advanced Settings")

        self._button_generate.on_click(self.generate_description)
        self._button_refresh.on_click(self._refresh_models_list)
        self._widget_source.observe(self._change_rate, names="value")
        self._widget_parent_select.observe(self._update_class_map)

        self._refresh()

        return self.kb_items



"""
ST-Link CLI Interface

The MIT License (MIT)

Copyright (c) 2016 mtchavez

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

Changes Made by SensiML corporation to create STLinkFlasher Class
Changes Copyright (c) 2019 SensiML Corporation
"""

import re
import subprocess
from typing import List
from pathlib import Path
from sensiml.device_flashing import DeviceFlasher


class STDeviceFlasher(DeviceFlasher):
    def __init__(self):
        super(STDeviceFlasher, self).__init__()
        self._name = "StLink Flasher"
        self._flash_methods = ["ST-Link SWD", "ST-Link JTAG"]
        self._supported_platforms = ["STMicro SensorTile"]
        self._needs_scan = True
        self._description = (
            "Flash Knowledge Pack binaries on STMicro Devices using ST-LINK_CLI"
        )

    @property
    def programmers(self):
        return self._programmers

    def flash(self, bin_file_path: Path, probe: int = 0, method: str = "SWD"):
        """
        Flashes microcontroller.

        The method flashes mcu using ST-Link Utility's CLI interface.
        To keep things simple most of the flags are hardcoded.

        Args:
            bin_file_path: Path of the hex file that will be used to flash the mcu
            probe: the probe number of the ST-Link programmer. Default is zero.
            method: Choose JTAG or SWD

        Returns:
            flash_status: Either 'successful' or 'failed'
            checksum: the hex file checksum provided by ST-Link Utility CLI.
                It will be zero if programming fails.
        """
        abs_path = bin_file_path.absolute()
        if "SWD" in method.upper():
            st_link_method = "SWD"
        elif "JTAG" in method.upper():
            st_link_method = "JTAG"
        print(method.lower())
        if "st-link" in method.lower():
            command = [
                "ST-LINK_CLI -c ID={0} {1} UR Hrst -Q -P {2} 0x08000000 -HardRST HIGH -Rst".format(
                    probe, st_link_method, abs_path
                )
            ]
            self._wait_for_process(command)



import sqlite3
import os
import pandas as pd
from sensiml.datasegments import DataSegments, DataSegment


class DCLProject:
    def __init__(self):
        self._path = None
        self._conn = None
        self._tables = None

    @property
    def data_dir(self):

        if self._path is None:
            raise Exception("Path is not set!")

        return os.path.join(self._path, "data")

    def _set_table_list(self):
        cursorObj = self._conn.cursor()

        cursorObj.execute('SELECT name from sqlite_master where type= "table"')

        self._tables = [x[0] for x in cursorObj.fetchall()]

    def create_connection(self, db_file):
        """ create a database connection to the SQLite database
            specified by db_file
        :param db_file: database file
        :return: Connection object or None
        """
        self._path = os.path.dirname(db_file)
        conn = None
        try:
            conn = sqlite3.connect(db_file)
        except Exception as e:
            print(e)

        self._conn = conn
        self._set_table_list()

    def _list_table(self, tablename):

        if tablename not in self._tables:
            print("Table is not part of the database.")
            return None
        cur = self._conn.cursor()
        cur.execute("SELECT * FROM {}".format(tablename))

        header = [x[0] for x in cur.description]
        rows = cur.fetchall()

        results = []
        for row in rows:
            results.append(row)

        return pd.DataFrame(results, columns=header)

    def list_captures(self):
        return self._list_table("Capture")

    def list_capture_labels(self, capture_name):
        cur = self._conn.cursor()
        cur.execute(
            'SELECT * FROM CaptureLabelValue JOIN Capture on CaptureLabelValue.capture=Capture.id where Capture.name="{}"'.format(
                capture_name
            )
        )

        header = [x[0] for x in cur.description]
        rows = cur.fetchall()

        results = []
        for row in rows:
            results.append(row)

        return pd.DataFrame(results, columns=header)

    def get_capture_segments(self, capture_name):

        labels = self.list_capture_labels(capture_name)

        tmp_df = pd.read_csv(
            os.path.join(self.data_dir, capture_name), index_col="sequence"
        )

        M = DataSegments()

        for index, label in enumerate(
            labels[
                ["capture_sample_sequence_start", "capture_sample_sequence_end"]
            ].values
        ):
            M.append(
                DataSegment(
                    tmp_df.loc[label[0] : label[1]],
                    metadata={"segment": index, "capture": capture_name},
                )
            )

        return M




import matplotlib.pyplot as plt


class DataSegment(object):
    def __init__(self, data, metadata, label=None):
        self._metadata = metadata
        self._label = label
        self._original_index = data.index
        self._data = data.reset_index(drop=True)

    def plot(self, **kwargs):
        self._data.plot(title=self.__str__(), **kwargs)

    def __str__(self):
        return " ".join(["{}:{}".format(k, v) for k, v in self._metadata.items()])


class DataSegments(list):
    def plot(self, **kwargs):
        for segment in self.__iter__():
            segment.plot(**kwargs)


def to_datasegments(data, metdata_columns, label_column):
    """ Converts a dataframe into a data setments """

    group_columns = metdata_columns + [label_column]
    g = data.groupby(group_columns)
    ds = DataSegments()

    data_columns = [x for x in data.columns if x not in group_columns]

    for key in g.groups.keys():

        metadata = {}
        for index, value in enumerate(group_columns):
            metadata[value] = key[index]

        tmp_df = g.get_group(key)[data_columns]

        ds.append(DataSegment(tmp_df, metadata, metadata[label_column]))

    return ds




from ipywidgets import HBox, VBox, Dropdown, Button, Image, Layout, Accordion
from numpy import mean, sort
from bqplot.marks import Graph, Image
from bqplot.scales import LinearScale, OrdinalScale
from bqplot.figure import Figure
from bqplot import Tooltip, ColorScale
from bqplot.axes import Axis


class HierarchicalModelVisualizationWidget:
    """
    Visualize hierarchical model tree.
    """

    def __init__(self, dsk=None, kp=None):
        self._dsk = dsk
        self.kp = kp
        self.plot_tree(kp)

    def find_deep(self, model_name):
        if model_name == "Parent":
            model_index = 0
        else:
            model_index = int(model_name.split("_")[1])

        height_of_tree = {
            "600": [0],
            "500": range(2, 4),
            "400": range(4, 8),
            "300": range(8, 16),
            "200": range(16, 32),
            "100": range(32, 64),
            "1": range(64, 128),
        }

        return int([h for h, v in height_of_tree.items() if model_index in v][0])

    def compute_x_axis(self, y_axis_loc, link_data):
        x_axis_location = {0: 3}
        distance = 0.64
        previous_source = y_axis_loc[0]
        x_axis_loc = [3]

        for i in link_data:
            target = i["target"]
            source = i["source"]

            if y_axis_loc[target] != previous_source:
                distance = distance * 0.5

            if (i["target"] % 2) == 0:
                x_axis_location[target] = x_axis_location[source] + distance
            else:
                x_axis_location[target] = x_axis_location[source] - distance

            x_axis_loc.append(x_axis_location[target])
            previous_source = y_axis_loc[target]

        return x_axis_loc

    def score_nodes(self, acc_dict, kb_description):
        label_list = list(acc_dict.keys() - kb_description.keys())
        while [label for label, value in acc_dict.items() if value is None]:

            for node in [label for label, value in acc_dict.items() if value is None]:
                both_node = (
                    sum(
                        [
                            label in kb_description.keys()
                            for _, label in kb_description[node]["results"].items()
                        ]
                    )
                    == 2
                )
                corrected_score = [
                    acc_dict[label]
                    for _, label in kb_description[node]["results"].items()
                ]

                # this is a node has 2 child nodes with corrected stat
                if both_node and None not in corrected_score:
                    acc_dict[node] = mean(
                        [
                            acc_dict[label]
                            for _, label in kb_description[node]["results"].items()
                        ]
                    )

                # this is a node has 1 child node and one leaf with corrected stat
                if (
                    sum(
                        [
                            label in label_list
                            for _, label in kb_description[node]["results"].items()
                        ]
                    )
                    == 1
                ):
                    acc_list = []
                    for _, label in kb_description[node]["results"].items():
                        if label in kb_description.keys():
                            acc_list.append(acc_dict[label])
                        else:
                            acc_list.append(acc_dict[label]["Accuracy"])

                    acc_dict[node] = mean(acc_list)

        return acc_dict

    def compute_hm_graph_paramenter(self, kp):
        kb_description = kp.knowledgepack_description
        kb_description_list = list(kb_description.keys())

        link_data = []
        colors = []
        node_names = []
        y_axis_loc = []
        num = 1
        link_map = {"Parent": 0}
        acc_dict = {}

        for i, node in enumerate(kb_description_list):
            node_height = self.find_deep(node)
            y_axis_loc.append(node_height)
            uuid = kb_description[node]["uuid"]
            temp_kp = self._dsk.get_knowledgepack(uuid)
            class_map = temp_kp.class_map
            acc_dict[node] = None

            if node == "Parent":
                hm_confusionmatrix = temp_kp.model_results["metrics"]["hierarchical"][
                    "ConfusionMatrix"
                ]

            child_nodes = kb_description[node]["results"]
            if node not in node_names:
                node_names.append(node)
                colors.append("#1f77b4")

            min_dec = 1
            leaf_acc_list = []
            for cn_index, cn_node in child_nodes.items():
                if cn_node != "Report":
                    if cn_node not in node_names:
                        node_names.append(cn_node)
                        colors.append("#1f77b4")
                        link_map[cn_node] = num
                else:
                    colors.append("#d62728")
                    y_axis_loc.append(node_height - 100)
                    class_label = class_map[cn_index]
                    class_acc = round(
                        100
                        * hm_confusionmatrix[class_label][class_label]
                        / sum(list(hm_confusionmatrix[class_label].values())),
                        2,
                    )
                    leaf_acc_list.append(class_acc)
                    acc_dict[class_label] = {
                        "Accuracy": class_acc,
                        "Predictions": hm_confusionmatrix[class_label],
                    }
                    kb_description[node]["results"][cn_index] = class_label
                    node_names.append(class_label)

                # This one compute acc scores for nodes that has 2 leafs
                if len(leaf_acc_list) == 2:
                    acc_dict[node] = mean(leaf_acc_list)

                link_data.append({"source": link_map[node], "target": num, "value": 1})
                num = num + 1

        y_axis_loc = list(sort(y_axis_loc)[::-1])
        shape_attrs = {"rx": 10, "ry": 10, "width": 60}
        node_data = []
        acc_dict = self.score_nodes(acc_dict, kb_description)

        for node_name in node_names:
            node_data_dict = {
                "label": node_name,
                "shape": "rect",
                "shape_attrs": shape_attrs,
                "Model": node_name,
                "Accuracy": str(acc_dict[node_name]),
            }
            node_data.append(node_data_dict)

        x_axis_loc = self.compute_x_axis(y_axis_loc, link_data)
        # compute the average accuracy of all leafs
        class_labels = list(acc_dict - kp.knowledgepack_description.keys())
        over_all_accuracy = round(
            mean([acc_dict[node]["Accuracy"] for node in class_labels]), 2
        )

        return (
            node_names,
            link_data,
            node_data,
            y_axis_loc,
            x_axis_loc,
            colors,
            over_all_accuracy,
        )

    def plot_tree(self, kp):
        node_names, link_data, node_data, y_axis_loc, x_axis_loc, colors, over_all_accuracy = self.compute_hm_graph_paramenter(
            kp
        )
        fig_layout = Layout(width="1024px", height="500px")
        graph = Graph(
            node_data=node_data,
            link_data=link_data,
            link_type="line",
            colors=colors,
            directed=False,
            scales={
                "x": LinearScale(),
                "y": LinearScale(),
                "link_color": ColorScale(scheme="Greens"),
            },
            x=x_axis_loc,
            y=y_axis_loc,
        )

        tooltip = Tooltip(fields=["Model", "Accuracy"], formats=["", ""])
        graph.tooltip = tooltip

        display(
            Figure(
                marks=[graph],
                layout=fig_layout,
                title="Hierarchical Model Tree Accuracy: " + str(over_all_accuracy),
            )
        )




from sensiml.datamanager.pipeline import PipelineStep


class TrainAndValidationCall(PipelineStep):
    """The base class for a train-and-validation call."""

    def __init__(self, name=""):
        super(TrainAndValidationCall, self).__init__(
            name=name, step_type="TrainAndValidationCall"
        )
        self._validation_methods = []
        self._classifiers = []
        self._optimizers = []
        self._input_data = ""
        self._feature_table = ""
        self._label_column = None
        self._ignore_columns = None
        self._outputs = []
        self._validation_seed = None

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        self._name = value

    @property
    def validation_methods(self):
        return self._validation_methods

    @validation_methods.setter
    def validation_methods(self, value):
        self._validation_methods = value

    @property
    def classifiers(self):
        return self._classifiers

    @classifiers.setter
    def classifiers(self, value):
        self._classfiers = value

    @property
    def optimizers(self):
        return self._optimizers

    @optimizers.setter
    def optimizers(self, value):
        self._optimizers = value

    @property
    def input_data(self):
        return self._input_data

    @input_data.setter
    def input_data(self, value):
        self._input_data = value

    @property
    def feature_table(self):
        return self._feature_table

    @feature_table.setter
    def feature_table(self, value):
        self._feature_table = value

    @property
    def label_column(self):
        return self._label_column

    @label_column.setter
    def label_column(self, value):
        self._label_column = value

    @property
    def ignore_columns(self):
        return self._ignore_columns

    @ignore_columns.setter
    def ignore_columns(self, value):
        self._ignore_columns = value

    @property
    def outputs(self):
        return self._outputs

    @outputs.setter
    def outputs(self, value):
        self._outputs = value

    @property
    def validation_seed(self):
        return self._validation_seed

    @validation_seed.setter
    def validation_seed(self, value):
        self._validation_seed = value

    def add_validation_method(self, validation_method_call):
        """Adds a validation method call to the train-and-validation call.

            Args:
                validation_method_call (ValidationMethodCall): object to append
        """
        self._validation_methods.append(validation_method_call)

    def remove_validation_method(self, validation_method_call):
        """Removes a validation method call from the train-and-validation call.

            Args:
                validation_method_call (ValidationMethodCall): object to remove
        """
        self._validation_methods = [
            f for f in self._validation_methods if f != validation_method_call
        ]

    def add_classifier(self, classifier_call):
        """Adds a classifier call to the train-and-validation call.

            Args:
                classifier_call (ClassifierCall): object to append
        """
        self._classifiers.append(classifier_call)

    def remove_classifier(self, classifier_call):
        """Removes a classifier call from the train-and-validation call.

            Args:
                classifier_call (ClassifierCall): object to remove
        """
        self._classifiers = [f for f in self._classifiers if f != classifier_call]

    def add_optimizer(self, optimizer_call):
        """Adds an optimizer call to the train-and-validation call.

            Args:
                optimizer_call (OptimizerCall): object to append
        """
        self._optimizers.append(optimizer_call)

    def remove_optimizer(self, optimizer_call):
        """Removes an optimizer call from the train-and-validation call.

            Args:
                optimizer_call (OptimizerCall): object to remove
        """
        self._optimizers = [f for f in self._optimizers if f != optimizer_call]

    def _to_dict(self):
        d = {}
        d["name"] = self.name
        d["type"] = "tvo"
        d["input_data"] = self.input_data
        d["feature_table"] = self.feature_table
        d["label_column"] = self.label_column
        d["ignore_columns"] = self.ignore_columns
        d["outputs"] = self._outputs
        d["validation_seed"] = self._validation_seed

        d["optimizers"] = []
        for item in self._optimizers:
            d["optimizers"].append(item._to_dict())

        d["classifiers"] = []
        for item in self._classifiers:
            d["classifiers"].append(item._to_dict())

        d["validation_methods"] = []
        for item in self._validation_methods:
            d["validation_methods"].append(item._to_dict())

        return d



import matplotlib.pyplot as plt


class VisualizeMixin:
    def plot_training_metrics(self, figsize=(16, 4)):
        """ plots the loss and validation accuracy during training for a model """
        if self.training_metrics:
            fig = plt.figure(figsize=figsize)
            num_iterations = len(self.training_metrics["validation_accuracy"])
            plt.plot(self.training_metrics["train_accuracy"], label="training")
            plt.plot(self.training_metrics["validation_accuracy"], label="validation")
            plt.xlabel("epochs")
            plt.ylabel("accuracy")
            plt.xlim(0, num_iterations)
            plt.ylim(0, 1)
            plt.legend(loc="best")
            plt.show()
        else:
            print("No training metrics for this model")



import os
import json
from pandas import DataFrame, Series

from sensiml.base import utility


class PlatformDescription(object):
    """Base class for a PlatformDescription object"""

    _id = 0
    _board_name = ""
    _hardware_accelerators = dict()
    _manufacturer = ""
    _can_build_binary = False
    _platform = ""
    _platform_version = ""
    _description = ""
    _supported_source_drivers = {}
    _supported_outputs = {}
    _ota_capable = False
    _target_os_options = {}
    _execution_parameters = {}

    def __init__(self, connection):
        self._connection = connection

    @property
    def board_name(self):
        return self._board_name

    @property
    def manufacturer(self):
        return self._manufacturer

    @property
    def id(self):
        return self._id

    @property
    def hardware_accelerators(self):
        return self._hardware_accelerators

    @property
    def can_build_binary(self):
        return self._can_build_binary

    @property
    def platform(self):
        return self._platform

    @property
    def description(self):
        return self._description

    @description.setter
    def description(self, value):
        self._description = value

    @property
    def platform_version(self):
        return self._platform_version

    @property
    def ota_capable(self):
        return self._ota_capable

    @property
    def supported_outputs(self):
        return self._supported_outputs

    @property
    def supported_source_drivers(self):
        return self._supported_source_drivers

    @property
    def target_os_options(self):
        return self._target_os_options

    @property
    def execution_parameters(self):
        return self._execution_parameters

    def get_config(self, test_data="", debug=False):
        """ Generates a default configuation dictionary for downloading a knowledgepack using this platform progrmatically

            Args:
                test_data (str): name of test data file to load onto knowledge pack,
                debug (bool): Build option for knowledge pack. Debug mode will have extra printouts to help with debugging issues.

        """

        if self.supported_source_drivers:
            source = self.supported_source_drivers["Default"][0]
        else:
            source = "Custom"

        if self.supported_outputs:
            outputs = [x.lower() for x in self.supported_outputs]
        else:
            outputs = ["serial"]

        config = {
            "target_platform": self.id,
            "test_data": test_data,
            "debug": debug,
            "application": self.execution_parameters.get("application", "Default"),
            "output_options": outputs,
            "sample_rate": self.supported_source_drivers.get("Default", [None, ""])[1],
            "source": source,
        }

        return config

    def refresh(self):
        """Calls the REST API and populates the local object properties from the server."""
        url = "platforms/{0}/".format(self.id)
        response = self._connection.request("get", url)
        response_data, err = utility.check_server_response(response)
        if err is False:
            self._board_name = response_data["board_name"]
            self._hardware_accelerators = response_data["hardware_accelerators"]
            self._platform = response_data["platform"]
            self._platform_version = response_data["platform_version"]
            self._description = response_data["description"]
            self._can_build_binary = response_data["can_build_binary"]
            self._ota_capable = response_data["ota_capable"]
            self._supported_source_drivers = response_data["supported_source_drivers"]
            self._supported_outputs = response_data["supported_outputs"]
            self._target_os_options = response_data["target_os_options"]
            self._execution_parameters = response_data["execution_parameters"]

    def initialize_from_dict(self, input_dictionary):
        """Populates a single transform object from a dictionary of properties from the server.

            Args:
                input_dictionary (dict): containing uuid, type, subtype, name, function_in_file, description,
                input_contract, and subtype
        """
        self._id = input_dictionary["id"]
        self._board_name = input_dictionary["board_name"]
        self._hardware_accelerators = input_dictionary["hardware_accelerators"]
        self._platform = input_dictionary["platform"]
        self._platform_version = input_dictionary[u"platform_version"]
        self._description = input_dictionary["description"]
        self._can_build_binary = input_dictionary["can_build_binary"]
        self._ota_capable = input_dictionary.get("ota_capable", False)
        self._supported_source_drivers = input_dictionary.get(
            "supported_source_drivers",
            {
                "Motion": [200, 100, 50, 25],
                "Audio": [16000],
                "Default": ["Motion", 100],
                "Custom": [],
            },
        )
        self._supported_outputs = input_dictionary.get(
            "supported_outputs", ["BLE", "SERIAL", "LED"]
        )
        self._target_os_options = input_dictionary.get(
            "target_os_options", ["ARM GCC Generic"]
        )
        self._execution_parameters = input_dictionary.get(
            "execution_parameters", {"application": "Default"}
        )

    def __dict__(self):
        ret = {
            "Id": int(self.id),
            "Board": self.board_name,
            "Hardware Accelerated": "Yes"
            if len(self.hardware_accelerators.items()) > 0
            else "No",
            "Software Platform": self.platform,
            "Platform Version": self.platform_version,
            "Description": self.description,
        }
        return ret

    def __call__(self):
        pd_dict = self.__dict__()
        pseries = Series(pd_dict, index=pd_dict.keys())
        df = DataFrame()
        df = df.append(pseries, ignore_index=True)
        return df.drop(labels=["Id"], axis=1)



try:
    from http.cookiejar import CookiePolicy
except ImportError:
    from cookielib import CookiePolicy
import os
import re
import json as JSON
import requests_oauthlib
import requests
from requests import Response
from oauthlib.oauth2 import (
    LegacyApplicationClient,
    InvalidGrantError,
    MissingTokenError,
    TokenExpiredError,
)
from requests_oauthlib import OAuth2Session, OAuth2
from getpass import getpass
from appdirs import user_config_dir
import errno
from uuid import uuid4
from time import sleep
import logging
import warnings
from functools import partial
from .connection_config import ConnectionConfig
from six.moves import input

try:
    import configparser as ConfigParser
except:
    import ConfigParser
from .errors import *

logger = logging.getLogger(__name__)

# Get configuration directory
config_dir = user_config_dir(__name__.split(".")[0], False)

cid = "f7J0QNKkKcIkHMpk5lvq0wGZtBEpw1YcJ8Bea9Pv"
csid = "4OxQzkEnVc6OmTMKHgi77o8uFhEDz1QCuyKsNFm2Jo84r0TlGBXli0NrZ5uHTI769P3UurvDVB86awZkrEikCUJ7RM333BTS5oajO32BRwsNkPcOzA3JlBkPblvKM0l0"


class Warnings:
    insecure_warning = "You have chosen to disable certificate verification. Your connection is not secure!"


class Connection(object):
    """Connection class"""

    def __init__(self, server=None, username=None, password=None, **kwargs):
        """Create Connection object

        Args:
            server (str, optional): Server URL or name in config file.
            username (str, optional)
            password (str, optional)
            **kwargs (optional): Mostly used when passing `server` as URL. See below.

        Keyword Args:
            insecure (bool): Whether or not to require valid SSL/TLS in requests
            path (str, optional): When specifiying a server name instead of a URL, the path of the config file.
            server_name (str, optional): user-readable name of server. defaults to the server domain.
            url (str, optional): Alternative name for server argument.
            auth_url (str, optional): URL of the oauth server. Defaults to server url.
        """
        # Check url parameter to allow direct passing of ConnectionConfig instances

        server = server or kwargs.get("url")
        if re.match(r"^https?://", server):
            self.server_url = re.sub(r"([^/])$", r"\1/", server)
            self.server_name = re.sub(r"^https?://([A-Za-z0-9.\-_]+).*$", r"\1", server)
            self.auth_url = kwargs.pop("auth_url", self.server_url)
            self.client_id = cid
            self.client_secret = csid
            self.username = username
        else:
            path = kwargs.pop("path", "connect.cfg")
            config = ConnectionConfig(server)
            try:
                config.load_from_file(path)
            except ConfigParser.NoSectionError:
                choice = (
                    input(
                        "No configuration found, would you like to create one? (y)/n "
                    )
                    or "y"
                )
                if choice.lower() in ["y", "yes", "t", "true"]:
                    config.create_config_object(path)
                else:
                    raise ConnectionAbortedError()

            self.server_name = config.server
            self.server_url = config.url
            self.auth_url = config.auth_url
            self.client_id = cid
            self.client_secret = csid
            self.username = username

        self.token_url = self.auth_url.rstrip("/") + "/token/"
        self.max_retries = kwargs.pop("max_retries", 3)

        self._token_filename = os.path.join(
            config_dir, "{}_token.json".format(self.server_name)
        )

        try:
            self._token_last_updated = os.path.getmtime(self._token_filename)
        except OSError as e:
            if e.errno == errno.ENOENT:
                self._token_last_updated = None
            else:
                raise

        # Disable SSL check for localhost or when insecure=True
        self.insecure = kwargs.pop("insecure", False)
        if self.insecure:
            logger.error(Warnings.insecure_warning)
        if re.match(r"^https?://localhost", self.server_url) or self.insecure:
            requests.packages.urllib3.disable_warnings()
            os.environ["OAUTHLIB_INSECURE_TRANSPORT"] = "1"
            self._verify = False
        else:
            os.environ["OAUTHLIB_INSECURE_TRANSPORT"] = ""
            self._verify = True

        self._login(password)

    def _login(self, password, attempt=0):
        if attempt >= self.max_retries:
            raise LoginAttemptsExceededError()

        self.session = OAuth2Session(
            client=LegacyApplicationClient(client_id=self.client_id),
            auto_refresh_url=self.token_url,
            token_updater=self._token_updater,
        )
        self._set_session_retries()
        self.session.cookies.set_policy(BlockAll())

        if self.username is None or password is None:
            if os.path.exists(self._token_filename):
                with open(self._token_filename, "r") as f:
                    self.session.token = JSON.load(f)
            else:
                self.username = input("Enter username: ")
                password = getpass("Enter password: ")

        if not self.session.token:
            try:
                self.session.fetch_token(
                    token_url=self.token_url,
                    username=self.username,
                    password=password,
                    # UAA endpoints always require id/secret as basic auth
                    auth=(self.client_id, self.client_secret),
                    verify=self._verify,
                )
            except (InvalidGrantError, MissingTokenError) as e:
                if isinstance(e, InvalidGrantError):
                    print("Invalid log in credentials, please try again.")
                else:
                    print(
                        "No token received: The remote server did not return a valid authorization response."
                    )
                if os.path.exists(self._token_filename):
                    os.remove(self._token_filename)
                password = None
                self.username = None
                return self._login(password, attempt=attempt + 1)
            self._save_token(self.session.token)
        logger.debug("Login completed")

    def request(self, method, path, json=None, headers=None, **kwargs):
        """Sends request to server with authentication

        Args:
            method (str): HTTP Method.
            path (str): Path from server root to request.
            json (dict, optional): dictionary of information to be submitted as json
            headers (dict, optional): http headers to be sent
            **kwargs: Additional keyword arguments passed to underlying request library.

        Keyword Args:
            data (dict): Data to upload as application/x-www-form-urlencoded or multipart/form-data.
                Do not use in conjunction with json.
            paged (boolean): Return [PagedRequest] iterator instead of response

        Returns:
            Response: Response object from requests library with the result of the request
        """

        attempt = kwargs.pop("attempt", 0)
        paged = kwargs.pop("paged", False)
        verify = kwargs.pop("verify", self._verify)
        url = self.server_url.strip("/") + "/" + path.lstrip("/")
        if self._token_last_updated != os.path.getmtime(self._token_filename):
            with open(self._token_filename, "r") as f:
                self.session.token = JSON.load(f)
        while attempt < self.max_retries:
            if paged:
                return PagedRequest(
                    self.session,
                    method,
                    url,
                    json=json,
                    headers=headers,
                    verify=verify,
                    client_id=self.client_id,
                    client_secret=self.client_secret,
                    **kwargs
                )
            else:
                response = self.session.request(
                    method,
                    url,
                    json=json,
                    headers=headers,
                    verify=verify,
                    client_id=self.client_id,
                    client_secret=self.client_secret,
                    **kwargs
                )
            # temp workaround for bug in django-oauth-toolkit
            if response.status_code not in [401]:
                return response
            attempt += 1
            try:
                # If normal autorefresh fails we force it
                token = self.session.refresh_token(
                    self.token_url,
                    auth=(self.client_id, self.client_secret),
                    verify=self._verify,
                )
                self._save_token(token)
            except (TokenExpiredError, InvalidGrantError) as e:
                print("Refresh Token expired, please re-enter your login information.")
                try:
                    os.unlink(
                        os.path.join(
                            config_dir, "{}_token.json".format(self.server_name)
                        )
                    )
                except OSError as e:
                    if e.errno != errno.ENOENT:
                        raise
                self._login(None)

        return response

    def paged_request(
        self, method, path, json=None, headers=None, limit=1000, **kwargs
    ):
        """Returns an iterator that will request a new page each call.

        The result returned by the iterator will be the json decoded data of the
        'results' property of the response. Function args are the same as [request()]

        Args:
            method
            path
            json
            headers
            limit (Optional[int]): Number of entries per page, defaults to 1000.
                Note: this will override the 'limit' param specified in 'params'
                request kwarg.
            **kwargs: passed to request method
        """
        params = kwargs.pop("params", {})
        params["limit"] = limit
        return self.request(
            method, path, json, headers, params=params, paged=True, **kwargs
        )

    def file_request(
        self, path, file_path, data, read_type="rb", method="post", **kwargs
    ):
        """Loads a file and formats metadata for specific endpoints and makes server request

        Args:
            path (str): Path to API endpoint
            file_path (str): Local path to file
            data (dict): Data to send to server along with request. May be sent in nonstandard headers instead of request body for some endpoints.
            method (str, optional): HTTP method, defaults to POST
            read_type (str, optional): Mode to open the file. defaults to binary read ('rb')

        Returns:
            Response
        """
        with open(file_path, read_type) as f:

            # TODO: Make this send params as body instead of headers.
            if path.find("segmenter/") >= 0:
                headers = {
                    "Name": data["name"],
                    "Type": data["specific_type"],
                    "Functioninfile": data["functioninfile"],
                }
                request = partial(self.request, method, path, data=f, headers=headers)

            # TODO: Make this send params as body instead of headers.
            elif path.find("transform/") >= 0:
                headers = {
                    "Name": data["name"],
                    "Type": data["type"],
                    "Functioninfile": data["functioninfile"],
                    "Subtype": data["subtype"],
                    "HasCVersion": data["has_c_version"],
                    "CFileName": data["c_file_name"],
                }
                request = partial(self.request, method, path, data=f, headers=headers)

            elif path.find("capture/") >= 0:
                request = partial(
                    self.request,
                    method,
                    path,
                    files={"file": f},
                    data={"name": data["name"], "asynchronous": data["asynchronous"]},
                )

            elif path.find("eventfile/") >= 0:
                request = partial(
                    self.request,
                    method,
                    path,
                    files={"data_file": f},
                    data={"name": data["name"]},
                )

            elif path.find("featurefile/") >= 0:
                request = partial(
                    self.request,
                    method,
                    path,
                    files={"data_file": f},
                    data={"name": data["name"]},
                )
            else:
                request = partial(
                    self.request, method, path, data=data, files={"file": f}
                )

            # call method with previously applied arguments
            return request(**kwargs)

    @staticmethod
    def logout(name):
        """Deletes saved session data for name"""
        try:
            os.unlink(os.path.join(config_dir, "{}_token.json".format(name)))
        except OSError as e:
            if e.errno == errno.ENOENT:
                print("No session saved for {}.".format(name))
                return
            else:
                raise
        print("Logged out.")

    def _set_session_retries(self):
        """set the number of retries a session has"""
        self.session.mount(
            "http://", requests.adapters.HTTPAdapter(max_retries=self.max_retries)
        )
        self.session.mount(
            "https://", requests.adapters.HTTPAdapter(max_retries=self.max_retries)
        )

    def _token_updater(self, token, save=True):
        """a simple function to save our token on refresh

        Args:
            token (dict): a token object for oauth2
        """
        self.session.token = token
        if save:
            self._save_token(token)

    def _save_token(self, token):
        if not os.path.exists(config_dir):
            try:
                os.makedirs(config_dir)
            except OSError as e:
                if e.errno != errno.EEXIST:
                    raise
        with open(self._token_filename, "w") as f:
            # f.write(json.dumps(token))
            JSON.dump(token, f)
            self._token_last_updated = os.path.getmtime(f.name)


# With ouath we don't want to send cookies, this will overwite and block session cookies
class BlockAll(CookiePolicy):
    return_ok = (
        set_ok
    ) = domain_return_ok = path_return_ok = lambda self, *args, **kwargs: False
    netscape = True
    rfc2965 = hide_cookie2 = False


class PagedRequest(object):
    def __init__(self, session, method, path, json=None, headers=None, **kwargs):
        self.session = session
        self._method = method
        self._path = path
        self._json = json
        self._headers = headers
        self._kwargs = kwargs

        self.response = None

    def __iter__(self):
        return self

    def next(self):
        """
        Returns:
            (dict) content of the results property in the response
        """
        if not self.response:
            self.response = self.session.request(
                self._method, self._path, self._json, self._headers, **self._kwargs
            )
        else:
            json = self.response.json()
            path = json["next"]
            if not path:
                raise StopIteration
            self.response = self.session.request(
                self._method, path, self._json, self._headers, **self._kwargs
            )
        self.response.raise_for_status()
        return self.response.json()["results"]




from ipywidgets import widgets
from ipywidgets import (
    Layout,
    Button,
    HBox,
    VBox,
    Accordion,
    FloatText,
    Textarea,
    Dropdown,
    Label,
    IntSlider,
    Checkbox,
    Text,
    Button,
    SelectMultiple,
    Password,
)
from IPython.display import display
from IPython.core.display import HTML as display_html
from ipywidgets import IntText
from json import dumps as jdump
from sensiml.widgets.base_widget import BaseWidget
from IPython.display import clear_output
from sensiml.widgets import (
    QueryWidget,
    AutoSenseWidget,
    DownloadWidget,
    FlashWidget,
    ModelExploreWidget,
)


layout = Layout(border="outset", width="100%")


class DashBoard(BaseWidget):
    def __init__(
        self,
        dsk=None,
        server="https://sensiml.cloud/",
        knowledgepack_level="pipeline",
        **kwargs
    ):
        self._server = server
        self._kwargs = kwargs
        self._dsk = dsk

        if self._dsk:
            self._dsk._auto_sense_ran = False
            self._dsk._query_created = False

        self._create_widget(knowledgepack_level=knowledgepack_level)

    def _set_project(self, option):

        if not isinstance(option["new"], dict):
            return
        if option["new"].get("index", None) is None:
            return

        project_name = option["owner"].options[option["new"]["index"]]

        if not project_name:
            return

        self._dsk.project = project_name

        pipelines = self._dsk.list_sandboxes()
        if pipelines is not None:
            pipelines = [""] + sorted(
                list(pipelines.Name.values), key=lambda s: s.lower()
            )
        else:
            pipelines = []

        self._widget_pipeline.options = pipelines
        self._widget_pipeline.observe(self._set_pipeline)
        self._refresh_project()

    def _set_pipeline(self, option):

        if not isinstance(option["new"], dict):
            return
        if option["new"].get("index", None) is None:
            return
        if not option["new"].get("index", None):
            return

        pipeline_name = option["owner"].options[option["new"]["index"]]
        self._dsk.pipeline = pipeline_name

        self._refresh_pipelines()

    def _logout(self, b):
        if self._dsk:
            self._dsk.logout()
            self._dsk = None
            self._clear_cell_output(None)

            self._widget_project.options = []
            self._widget_pipeline.options = []
            self._auto._clear()
            self._download._clear()
            self._explore._clear()
            self._button_connected.description = ""
            self._button_connected.icon = ""
            self._button_connected.style.button_color = "#00b300"
            self._button_connected.disabled = True

    def _clear_cell_output(self, b, return_display=True):
        clear_output(wait=True)

        display(display_html("<style>.container { width:90% !important; }</style>"))

        if return_display:
            return display(self.v)

    def _on_button_clicked(self, b):
        from sensiml import SensiML

        if self._dsk is None:
            try:
                self._dsk = SensiML(server=self._server)

                self._dsk._auto_sense_ran = False
                self._dsk._query_created = False
            except:
                print("Login Failed.")
                return

        self._clear_cell_output(None, return_display=False)

        self._button_connected.description = "Connected"
        self._button_connected.icon = "link"
        self._button_connected.style.button_color = "#0075c1"

        self._widget_project.options = [""] + sorted(
            list(self._dsk.list_projects().Name.values), key=lambda s: s.lower()
        )
        self._widget_project.observe(self._set_project)
        self._refresh_login()

    def _add_pipeline(self, b):

        if self._dsk is None:
            print("You must login first.")
            return
        if self._dsk.project is None:
            print("Project Must be set first.")
            return

        if not self._widget_new_pipeline.value:
            return None

        if self._widget_new_pipeline.value not in self._widget_pipeline.options:
            self._dsk.pipeline = self._widget_new_pipeline.value

        self._widget_pipeline.options = self._widget_pipeline.options + (
            self._widget_new_pipeline.value,
        )
        self._widget_pipeline.value = self._widget_new_pipeline.value

        self._refresh_pipelines()

    def _create_widget(self, knowledgepack_level="Project"):

        self._button_connected = Button(description="Connected")
        self._button_loggout = Button(description="Logout")
        self._button_clear = Button(
            description="Clear Log", tooltip="Clear Log Text Below Dashboard"
        )
        self._widget_project = Dropdown(description="Project", options=[])
        self._widget_pipeline = Dropdown(description="Pipeline", options=[])
        self._widget_new_pipeline = Text(description="New Pipeline", value="")
        self._button_add_pipeline = Button(description="Add", icon="plus")

        self._button_connected.on_click(self._on_button_clicked)
        self._button_clear.on_click(self._clear_cell_output)
        self._button_add_pipeline.on_click(self._add_pipeline)
        self._button_loggout.on_click(self._logout)

        login = VBox(
            [
                HBox(
                    [
                        VBox([self._button_connected, self._button_loggout]),
                        VBox([self._widget_project, self._widget_pipeline]),
                        VBox(
                            [
                                HBox(
                                    [
                                        self._widget_new_pipeline,
                                        self._button_add_pipeline,
                                    ]
                                )
                            ]
                        ),
                    ]
                )
            ],
            layout=layout,
        )

        bottom = VBox([self._button_clear], layout=layout)

        self._query = QueryWidget(None)
        self._auto = AutoSenseWidget(None, self._widget_project, self._widget_pipeline)
        self._download = DownloadWidget(None, level=knowledgepack_level)
        self._flash = FlashWidget(None)
        self._explore = ModelExploreWidget(None, level=knowledgepack_level)

        accordions = [
            Accordion(children=[self._query.create_widget()]),
            Accordion(children=[self._auto.create_widget()]),
            Accordion(children=[self._explore.create_widget()]),
            Accordion(children=[self._download.create_widget()]),
            Accordion(children=[self._flash.create_widget()]),
        ]
        accordions[0].set_title(0, "Data Exploration")
        accordions[1].set_title(0, "Model Building")
        accordions[2].set_title(0, "Explore Models")
        accordions[3].set_title(0, "Create Knowledge Pack")
        accordions[4].set_title(0, "Deploy Knowledge Pack")

        for a in accordions:
            a.selected_index = None

        accordions[1].selected_index = 0
        self.v = VBox([login] + accordions + [bottom])
        self.v.layout.border = "outset"

        self._on_button_clicked(None)

        if self._dsk:
            display(self.v)

    def _refresh_login(self):
        self._flash._dsk = self._dsk
        self._flash._refresh()

    def _refresh_project(self):
        self._query._dsk = self._dsk
        self._query._refresh()
        self._auto._clear()
        self._download._clear()
        self._explore._clear()

    def _refresh_pipelines(self):
        self._auto._dsk = self._dsk
        self._auto._refresh()
        self._download._dsk = self._dsk
        self._download._refresh()
        self._explore._dsk = self._dsk
        self._explore._refresh()



from ipywidgets import widgets
import qgrid
from ipywidgets import (
    Layout,
    Button,
    HBox,
    VBox,
    Box,
    HTML,
    FloatText,
    Tab,
    Textarea,
    Dropdown,
    Label,
    IntSlider,
    Checkbox,
    Text,
    Button,
    SelectMultiple,
)
from IPython.display import display
from ipywidgets import IntText
from sensiml.widgets.base_widget import BaseWidget, button_handling
from sensiml.base.exceptions import QueryExistsException
from six import text_type
import numpy as np
from IPython.display import display
from bqplot import OrdinalScale, LinearScale, Bars, Lines, Axis, Figure
from sensiml.widgets.renderers import WidgetAttributeRenderer
from pandas import DataFrame


def get_item(x, index, default=None):
    try:
        return x[index]
    except:
        return default


class QueryWidget:
    def __init__(self, dsk):
        self._dsk = dsk
        self._items = {"query_select": ""}
        self._stats = None

    def _enable_buttons(self):

        self._add_query.disabled = False
        self._update_query.disabled = False
        self._delete_query.disabled = False

    def _disable_buttons(self):

        self._add_query.disabled = True
        self._update_query.disabled = True
        self._delete_query.disabled = True

    @button_handling
    def on_add_query_button_clicked(self, b):

        if self.add_query(create=True, force=False):
            self.query_widget_tab.selected_index = 0

    @button_handling
    def on_update_query_button_clicked(self, b):
        if self._query_select.value:
            self.add_query(force=True)

    @button_handling
    def on_delete_query_button_clicked(self, b):
        if self._dsk is None:
            return
        if self._dsk.project is None:
            return
        if self._query_select.value:
            query = self._dsk.project.queries.get_query_by_name(
                self._query_select.value
            )
            query.delete(renderer=self.renderer)

        self._refresh()

    def add_query(self, create=False, force=False):
        if self._dsk is None:
            return
        if self._dsk.project is None:
            return
        if self._query_select.options == self._query_name_text.value and force == False:
            return

        self.populate_items(populate_create=create)
        query_name = self._items.get("query_name", None)
        segmenter_name = self._items.get("segmenter_name", None)
        label_column = self._items.get("label_column")
        metadata_columns = self._items.get("metadata_columns", None)
        sensor_columns = self._items.get("sensor_columns", None)
        metadata_filter = self._items.get("metadata_filter", "")

        if not segmenter_name:
            self.renderer.render("No segmenter specified")
            return

        # reverse lookup of segmenter id
        segmenters = self._dsk.list_segmenters()
        segmenter_id = int(
            segmenters[segmenters["name"] == self._items["segmenter_name"]][
                "id"
            ].values[0]
        )

        if not query_name:
            self.renderer.render("No query name specified")
            return

        if label_column in metadata_columns:
            self.renderer.render("Label Column cannot be part of metadata columns.")
            return

        if not metadata_columns:
            self.renderer.render("No Metadata Columns Specified")
            return
        if isinstance(metadata_columns, str):
            metadata_columns = metadata_columns.split(",")

        if not sensor_columns:
            self.renderer.render("No Sensor Columns Specified")
            return

        if isinstance(sensor_columns, str):
            sensor_columns = sensor_columns.split(",")

        try:
            self.renderer.render("Creating Query.")
            self.query = self._dsk.create_query(
                query_name,
                columns=sorted(sensor_columns),
                metadata_columns=[m for m in metadata_columns if m != label_column],
                segmenter=segmenter_id,
                metadata_filter=metadata_filter,
                label_column=label_column,
                force=force,
                renderer=self.renderer,
            )
            self.renderer.render("Query Created.")

        except QueryExistsException:
            self.renderer.render("Query Already Exists. Select force to overwrite.")
            return None

        if create:
            self._query_select.options = sorted(
                self._query_select.options + (query_name,), key=lambda s: s.lower()
            )
            self._query_select.value = query_name
            self._query_name_text.value = ""
            self._metadata_filter_text_create.value = ""
            self._segmenter_select_create.value = ""

        else:
            self._query_select.options = sorted(
                self._query_select.options, key=lambda s: s.lower()
            )
            self._query_select.value = query_name
            self.update_statistics(self.query)

        return True

    @button_handling
    def update_statistics(self, query=None, plot_type="segments"):

        if query is not None:
            try:
                self.renderer.render("Getting Query Statistics.")
                self._stats = query.statistics_segments(renderer=self.renderer)
                self.renderer.render("Revieved Query STatistics.")
                self.renderer.render("")

            except:
                self._graph.marks = []
                self._graph.axes = []
                self._graph.title = ""
                return

        else:
            query = self.query

        data = []
        Labels = []
        counts = []
        segments = []
        stacked_data = []

        g = self._stats.groupby(query.label_column)

        for tmp_df in g:
            Labels.append(tmp_df[0])
            data.append(sorted(list(tmp_df[1]["Segment Length"]), reverse=True))
            counts.append((tmp_df[1]["Segment Length"]).sum())
            segments.append(tmp_df[1]["Segment Length"].shape[0])

        num_classes = len(data)
        max_segments = max(segments)

        # Stacked bar Chart
        for j in range(max_segments):
            stacked_data.append([get_item(data[i], j, 0) for i in range(num_classes)])

        x_ord = OrdinalScale()
        y_lin = LinearScale(min=0.0)

        ax_x = Axis(scale=x_ord, tick_values=Labels, tick_rotate=10)
        ax_y = Axis(
            scale=y_lin, orientation="vertical", label=plot_type, label_offset="60px"
        )

        if plot_type == "samples":
            bar = Bars(
                x=Labels,
                y=counts,
                color_mode="group",
                type="stacked",
                scales={"x": x_ord, "y": y_lin},
            )

        elif plot_type == "segments":
            bar = Bars(
                x=Labels,
                y=segments,
                color_mode="group",
                type="stacked",
                scales={"x": x_ord, "y": y_lin},
            )

        elif plot_type == "samples_stacked":
            bar = Bars(
                x=Labels,
                y=stacked_data,
                color_mode="group",
                type="stacked",
                scales={"x": x_ord, "y": y_lin},
            )

        else:
            raise Exception("Plot Type Not Supported.")

        self._graph.marks = [bar]
        self._graph.axes = [ax_x, ax_y]
        self._graph.title = ""

    def _refresh(self):
        self._dsk.project.refresh()
        self._stats = None

        sensor_items = []
        metadata_items = []
        queries = []

        self._graph.marks = []
        self._graph.axes = []
        self._graph.title = ""

        segmenters = self._dsk.list_segmenters()
        if segmenters is None:
            self.renderer.render("No queryable data added to this project")
            segmenters = []
        else:
            segmenters = sorted(
                list(segmenters["name"].values), key=lambda s: s.lower()
            )

        columns = self._dsk.project.columns()
        if columns is None:
            self.renderer.render("No schema associated with this project.")
            columns = []

        for column in sorted(columns):
            sensor_items.append(column)

        metadata_items = self._dsk.project.metadata_columns()
        if metadata_items is None:
            self.renderer.render("No metdata associated with this project")
            metadata_items = []

        label_items = []
        for label in self._dsk.project.label_columns():
            if label != "SegmentID":
                label_items.append(label)

        query_df = self._dsk.list_queries()
        if query_df is not None:
            for query in query_df["Name"].values:
                queries.append(query)

        self._segmenter_select.options = [""] + segmenters
        self._segmenter_select_create.options = [""] + segmenters
        self._segmenter_select.value = self._segmenter_select.options[0]
        self._label_column_select.options = label_items
        self._label_column_select_create.options = label_items

        self._metadta_items_mselect.options = metadata_items
        self._metadata_items_mselect_create.options = metadata_items
        for name in metadata_items:
            if name.lower() in ["subject", "subjects"]:
                self._metadata_items_mselect_create.value = [name]

        self._sensor_columns_mselector.options = sensor_items
        self._sensor_columns_mselector_create.options = sensor_items
        self._sensor_columns_mselector_create.value = sensor_items
        self._query_select.options = [""] + sorted(
            list(set(queries)), key=lambda s: s.lower()
        )
        self._query_name_text.value = ""
        self._metadata_filter_text_create.value = ""
        self._metadata_filter_text.value = ""
        self.populate_items()

        project_stats, _ = self._dsk.project.statistics()
        self._w_project_explorer.df = project_stats

        return True

    def populate_items(self, populate_create=False):

        self._items["sensor_columns"] = self._sensor_columns_mselector.value
        self._items["metadata_columns"] = self._metadta_items_mselect.value
        self._items["label_column"] = self._label_column_select.value
        self._items["segmenter_name"] = self._segmenter_select.value
        self._items["query_name"] = self._query_select.value
        self._items["query_select"] = self._query_select.value
        self._items["metadata_filter"] = self._metadata_filter_text.value

        if populate_create:
            self._items["sensor_columns"] = self._sensor_columns_mselector_create.value
            self._items["metadata_columns"] = self._metadata_items_mselect_create.value
            self._items["label_column"] = self._label_column_select_create.value
            self._items["segmenter_name"] = self._segmenter_select_create.value
            self._items["query_name"] = self._query_name_text.value
            self._items["metadata_filter"] = self._metadata_filter_text_create.value

    def populate_query(self, change=None):
        self.populate_items()

        self.query = self._dsk.project.queries.get_query_by_name(
            self._items["query_select"]
        )

        if self.query:
            qdict = self.query._to_dict()
            self._qdict = qdict
            seg = self._dsk.list_segmenters()

            self._sensor_columns_mselector.value = qdict["sensor_columns"]
            self._metadta_items_mselect.value = qdict["metadata_columns"]
            self._label_column_select.value = qdict["label_column"]
            self._metadata_filter_text.value = qdict["metadata_filter"]

            if list(seg[seg["id"] == qdict["segmenter_id"]]["name"]):
                self._segmenter_select.value = list(
                    seg[seg["id"] == qdict["segmenter_id"]]["name"]
                )[0]
            else:
                self.renderer.render(
                    "Error. Query {} is associated with a Segmenter that no longer exists!".format(
                        self._items["query_select"]
                    )
                )
                return

            self.populate_items()

            self.update_statistics(self.query)

    def switch_plots(self, change=None):

        if self._stats is not None:
            self.update_statistics(None, change["new"])

    def create_widget(self, name=""):
        # refresh the project so that the most up to date values are shown

        self._query_name_text = Text(description="Name")
        self._query_select = Dropdown(description="Name")
        self._segmenter_select = Dropdown(description="Session")
        self._plot_type = Dropdown(
            description="Plot",
            options=["segments", "samples", "samples_stacked"],
            value="segments",
        )
        self._segmenter_select_create = Dropdown(description="Session")
        self._label_column_select = Dropdown(description="Label")
        self._label_column_select_create = Dropdown(description="Label")
        self._metadta_items_mselect = SelectMultiple(description="Metadata")
        self._metadata_items_mselect_create = SelectMultiple(description="Metadata")
        self._sensor_columns_mselector = SelectMultiple(description="Sources")
        self._sensor_columns_mselector_create = SelectMultiple(description="Sources")
        self._metadata_filter_text = Text(description="Query Filter")
        self._metadata_filter_text_create = Text(description="Query Filter")
        self._add_query = Button(
            icon="plus",
            description="Add",
            tooltip="Create New Query",
            layout=Layout(width="98%", allign_item="center"),
        )
        self._update_query = Button(
            icon="upload",
            description="Update",
            tooltip="Update Query",
            layout=Layout(width="98%", allign_item="center"),
        )
        self._delete_query = Button(
            icon="trash",
            description="Delete",
            tooltip="Delete Query",
            layout=Layout(width="98%", allign_item="center"),
        )

        self._w_project_explorer = qgrid.show_grid(
            DataFrame({"capture_name": [""]}),
            grid_options={
                "rowHeight": 32,
                "maxVisibleRows": 30,
                "minVisibleRows": 6,
                "editable": False,
                "defaultColumnWidth": 100,
                "forceFitColumns": True,
                "sortable": True,
                "filterable": True,
            },
            column_options={"minWidth": 30},
            column_definitions={"index": {"width": 30}, "capture_name": {"width": 350}},
        )

        self._add_query.on_click(self.on_add_query_button_clicked)
        self._update_query.on_click(self.on_update_query_button_clicked)
        self._delete_query.on_click(self.on_delete_query_button_clicked)

        self._graph = Figure(
            marks=[],
            axes=[],
            title="",
            legend_location="bottom-right",
            background_style={"fill": "white"},
            fig_margin={"top": 10, "bottom": 30, "left": 100, "right": 0},
        )
        self._graph.layout.height = text_type("325px")
        self._graph.layout.width = "70%"

        self._query_select.observe(self.populate_query, "value")
        self._plot_type.observe(self.switch_plots, "value")

        self._widget_render_space = HTML()

        self.renderer = WidgetAttributeRenderer(self._widget_render_space, "value")

        self.query_widget_tab = Tab(
            [
                HBox(
                    [
                        VBox(
                            [
                                self._query_select,
                                self._segmenter_select,
                                self._label_column_select,
                                self._metadta_items_mselect,
                                self._sensor_columns_mselector,
                                self._metadata_filter_text,
                                self._plot_type,
                                HBox([self._delete_query, self._update_query]),
                            ]
                        ),
                        self._graph,
                    ]
                ),
                HBox(
                    [
                        VBox(
                            [
                                self._query_name_text,
                                self._segmenter_select_create,
                                self._label_column_select_create,
                                self._metadata_items_mselect_create,
                                self._sensor_columns_mselector_create,
                                self._metadata_filter_text_create,
                                self._add_query,
                            ]
                        )
                    ]
                ),
                HBox([self._w_project_explorer]),
            ]
        )

        self.query_widget = VBox([self.query_widget_tab, self._widget_render_space])

        self.query_widget_tab.set_title(0, "Select/Modify Query")
        self.query_widget_tab.set_title(1, "Create New Query")
        self.query_widget_tab.set_title(2, "Project Explorer")

        if self._dsk:
            if self._refresh():
                self.populate_query()

        return self.query_widget



from sensiml.device_flashing import DeviceFlasher
from pathlib import Path


class NordicThingyFlasher(DeviceFlasher):
    def __init__(self):
        super(NordicThingyFlasher, self).__init__()
        self._name = "Nordic nrfjprog flasher"
        self._flash_methods = ["J-Link"]
        self._programmers = None
        self._supported_platforms = ["Nordic Thingy"]
        self._needs_scan = False
        self._description = "Flash Knowledge Pack binaries on Nordic devices using nrfutils' nrfjprog J-Link utility."
        self._flash_file_ext = ".hex"

    @property
    def programmers(self):
        return self._programmers

    @staticmethod
    def refresh_programmers_list():
        print(
            "Nordic programmer uses NRFJprog, and does not need to scan for programmers."
        )

    @staticmethod
    def _find_all_probes():
        return list()

    def flash(self, hex_file_path: Path, probe: int = 0, method: str = ""):
        """
        Flashes microcontroller.

        The method flashes mcu using ST-Link Utility's CLI interface.
        To keep things simple most of the flags are hardcoded.

        Args:
            hex_file_path: Path of the hex file that will be used to flash the mcu
            probe: unused for Nordic

        Returns:
            flash_status: Either 'successful' or 'failed'
            checksum: the hex file checksum provided by ST-Link Utility CLI.
                It will be zero if programming fails.
        """
        abs_path = hex_file_path.absolute()
        programming_command = [
            "nrfjprog --program {} -f nrf52 --sectorerase && nrfjprog --reset -f nrf52".format(
                abs_path
            )
        ]
        self._wait_for_process(programming_command)



from sensiml.method_calls.basemethodcall import BaseMethodCall


class OptimizerCall(BaseMethodCall):
    """The base class for an optimizer call.

    Child classes have their own additional properties and overwrite this docstring."""

    def __init__(self, name, function_type="optimizer"):
        super(OptimizerCall, self).__init__(name=name, function_type=function_type)
        self._refinement = {}



from sensiml.method_calls.basemethodcall import BaseMethodCall


class TrainingAlgorithmCall(BaseMethodCall):
    """The base class for a training algorithm call.

    Child classes have their own additional properties and overwrite this docstring."""

    def __init__(self, name, function_type="Training Algorithm"):
        super(TrainingAlgorithmCall, self).__init__(
            name=name, function_type=function_type
        )
        self._refinement = {}



import os
import sys
import io
import json
from ast import literal_eval
import IPython
from pandas import DataFrame, read_csv, concat
from ipywidgets import widgets
from ipywidgets import (
    Layout,
    Tab,
    BoundedIntText,
    Button,
    VBox,
    HBox,
    Box,
    FloatText,
    Textarea,
    Dropdown,
    Label,
    IntSlider,
    Checkbox,
    Text,
    Image,
    Button,
    SelectMultiple,
    HTML,
)
from IPython.display import display
from ipywidgets import IntText
from json import dumps as jdump
from sensiml.widgets.base_widget import BaseWidget, button_handling
from sensiml.widgets.renderers import WidgetAttributeRenderer
from copy import deepcopy
import numpy as np
import matplotlib.pyplot as plt
import qgrid

category_item_layout = Layout(
    # display='flex',
    size=16,
    # border='solid 2px',
    justify_content="flex-start",
    # background_color= 'red',
    overflow="visible",
)


def clean_name(name):
    return "".join(e if e.isalnum() else "_" for e in name)


class ModelExploreWidget(BaseWidget):
    def __init__(self, dsk=None, level="Project", folder="knowledgepacks"):
        self._dsk = dsk
        self.kp = None
        if not os.path.exists(folder):
            os.makedirs(folder)
        self.setup(level=level)
        self._last_capture_file = None
        self._last_parent_value = None

    def _enable_buttons(self):

        self._button_recognize_signal_capturefile.disabled = False

    def _disable_buttons(self):

        self._button_recognize_signal_capturefile.disabled = True

    def setup(self, level):
        self.level = level

    def _find_child_models(self, kps):
        child_uuid = []
        for index in kps[~kps.kp_description.isna()].index:
            knowledgepack_description = kps.loc[index, "kp_description"]
            child_uuid.extend(
                [
                    knowledgepack_description[child_model]["uuid"]
                    for child_model in list(
                        knowledgepack_description.keys() - set(["Parent"])
                    )
                ]
            )

        return kps[~kps.kp_uuid.isin(child_uuid)].reset_index(drop=True)

    def get_kp_dict(self):
        if self.level.lower() == "project":
            kps = self._dsk.project.list_knowledgepacks()
        elif self.level.lower() == "pipeline":
            kps = self._dsk.pipeline.list_knowledgepacks()
        else:
            kps = self._dsk.list_knowledgepacks()

        if isinstance(kps, DataFrame) and len(kps):
            kps = self._find_child_models(kps)
            kps = sorted(
                [
                    (name, value)
                    for name, value in kps[["Name", "kp_uuid"]].values
                    if name
                ],
                key=lambda s: s[0].lower(),
            )
            return kps

        return [("", None)]

    def _load_results(self, uuid, file_name):

        model_results_folder = "model_results"
        kp_folder = os.path.join(model_results_folder, uuid)
        results_file_path = os.path.join(kp_folder, file_name)

        if not os.path.exists(model_results_folder):
            os.mkdir(model_results_folder)
            return None

        if not os.path.exists(kp_folder):
            os.mkdir(kp_folder)
            return None

        if not os.path.exists(results_file_path):
            return None

        return read_csv(results_file_path)

    def _cache_results(self, uuid, file_name, results):

        model_results_folder = "model_results"
        kp_folder = os.path.join(model_results_folder, uuid)
        results_file_path = os.path.join(kp_folder, file_name)

        if not os.path.exists(model_results_folder):
            os.mkdir(model_results_folder)

        if not os.path.exists(kp_folder):
            os.mkdir(kp_folder)

        return results.to_csv(results_file_path)

    def _clear_the_cache(self, b, recognize=True):
        '''
            Delete the cache data for selected file
        '''

        name = "Name"
        if self._w_file_list.value == "captures":
            name = "capture_name"

        if self.kp is None:
            self._widget_render_space.value = "Please select a model."
        else:
            capture = None
            for capture in self._widget_capture_data.get_selected_df()[name].values:
                model_results_folder = "model_results"
                kp_folder = os.path.join(model_results_folder, self.kp.uuid )
                results_file_path = os.path.join(kp_folder, capture)
                
                if os.path.exists(results_file_path):
                    os.remove(results_file_path)
                    self._widget_render_space.value = "The cache is cleaned for the model: " + self.kp.name + " and the file: "  + capture
                    
                else:
                    self._widget_render_space.value = "There is no cached data for the model: " + self.kp.name + " and the file: "  + capture
                
            if capture is None:
                self._widget_render_space.value = "Please select the file to clean the cache." 
            
    def _recognize_signal_all(self, b, recognize=True):

        name = "Name"
        if self._w_file_list.value == "captures":
            name = "capture_name"

        if self.kp is None:
            self._widget_render_space.value = "Please select a model."
        else:
            capture = None
            for capture in self._widget_capture_data.get_selected_df()[name].values:
                self._button_clear_the_cache.disabled = True
                self._recognize_signal(capture)
                self._button_clear_the_cache.disabled = False

            if capture is None:
                self._widget_render_space.value = "Please select an input file."
    
    @button_handling
    def _recognize_signal(self, capture_file, recognize=True):

        self._last_capture_file = capture_file

        self._output_recognition_classification.layout.visibility = "hidden"

        if not capture_file or not self.kp.uuid:
            return

        if self._widget_parent_select.value != self.kp.uuid:
            return

        self.results = self._load_results(self.kp.uuid, capture_file)

        if self.results is None and not recognize:
            return None

        if self.results is None:
            self._button_recognize_signal_capturefile.description = (
                "Classifying Signal..."
            )

            if self.kp.knowledgepack_description:
                kp_description = deepcopy(self.kp.knowledgepack_description)
                del kp_description["Parent"]["source"]
            else:
                kp_description = None

            if self._w_file_list.value == "captures":
                self.results, s = self.kp.recognize_signal(
                    capture=capture_file,
                    renderer=self.renderer,
                    kb_description=kp_description,
                )
            else:
                self.results, s = self.kp.recognize_signal(
                    datafile=capture_file,
                    renderer=self.renderer,
                    kb_description=kp_description,
                )

            self._button_recognize_signal_capturefile.description = "RUN"

            if not isinstance(self.results, DataFrame):
                return None

            self._cache_results(self.kp.uuid, capture_file, self.results)

        if self.kp.knowledgepack_description:
            ytick_labels = ["Unknown"] + list(
                np.unique(self.kp.model_results["metrics"]["hierarchical"]["y_true"])
            )
            # update Classification index to flat model index
            self.results.Classification = [
                ytick_labels.index(i) for i in self.results.ClassificationName
            ]
        else:
            ytick_labels = ["Unknown"] + [
                self.kp.class_map[x] for x in sorted(self.kp.class_map, key=lambda x: int(x))
            ]

        self.results = post_process(self.results, self._w_post_process_steps.value)

        if self.results.shape[0] == 0:
            err = "Post Processing buffer is larger than the number of classifications from capture {}".format(
                capture_file
            )
            self._widget_render_space.value = err
            print(err)
            return

        if self._check_show_features.value:
            fig = get_classification_and_features_plot(
                self.kp,
                capture_file,
                self.results,
                self._w_post_process_steps.value,
                ytick_labels,
                show_numbers=self._check_show_feature_values.value,
            )
        else:
            fig = get_classification_plot(
                self.kp,
                capture_file,
                self.results,
                self._w_post_process_steps.value,
                ytick_labels,
            )

        buf = io.BytesIO()
        fig.savefig(buf, format="png")
        buf.seek(0)

        self._output_recognition_classification.value = buf.read()
        self._output_recognition_classification.layout.visibility = "visible"

        plt.show()

        """
        M = []
        for i in range(self.results.shape[0]):
            M.append(eval(self.results.loc[i]['FeatureVector']))
        columns = []
        for feature in self.kp.feature_summary:
            columns.append(feature['Feature'])
        fv = DataFrame(M, columns=columns)
        fv['Labels'] = self.results.ClassificationName

        # self._dsk.pipeline.visualize_features(
        #    fv, label_column="Labels")

        # self._output_recognition_features.value = buff.read()
        # self._output_recognition_features.visibility = "visible"
        """

    def _rehydrate_model(self, b):

        if self.kp:
            pipeline_steps = self._dsk.pipeline.rehydrate_knowledgepack(
                self.kp, replace=False
            )

            pipeline_steps.append("r,s = dsk.pipeline.execute()")

            print("\n\n".join(pipeline_steps))

            instructions = "#### This is a rehydrated Knowledge Pack. \n\n#### Important: Remove the quotes at the start and execute this cell to run. \n\n\n"

            df = DataFrame(
                [
                    '"\n'
                    + instructions
                    + "\n\n".join(pipeline_steps).replace('"', "'")
                    + '\n\n\n########""""""'
                ]
            )

            df.to_clipboard(index=False, header=False, excel=True)

            return pipeline_steps

        return ""

    def _refresh(self, b=None):
        if self._dsk is None:
            return

        self._refresh_file_list()

        self._widget_parent_select.options = self.get_kp_dict()
        self._widget_parent_select.value = self._widget_parent_select.options[0][1]

    def _refresh_file_list(self, b=None):
        if self._dsk is None:
            return

        if b is None or b["type"] == "change" and b["name"] == "value":
            if self._w_file_list.value == "captures":
                project_stats, _ = self._dsk.project.statistics()
            else:
                project_stats = self._dsk.list_datafiles()

                if not isinstance(project_stats, DataFrame):
                    project_stats = DataFrame({"Name": [""]})

            self._widget_capture_data.df = project_stats

    def _clear(self):
        self._widget_parent_select.options = [""]
        self._widget_parent_select.value = ""
        self._widget_confusion_matrix.value = ""
        self._widget_feature_summary.value = ""
        self._widget_model_summary.value = ""
        self._output_recognition_classification.layout.visibility = "hidden"
        self._widget_capture_data.df = DataFrame({"capture_name": [""]})

    def _update_confusion_matrix(self, *args):

        if self._last_parent_value == self._widget_parent_select.value:
            return

        if self._widget_parent_select.value:
            self._last_parent_value = self._widget_parent_select.value
            self.kp = self._dsk.get_knowledgepack(self._widget_parent_select.value)

            self._widget_confusion_matrix.value = self._get_confusion_matrix_html(
                self.kp
            )

            self._widget_feature_summary.value = self._get_feature_summary_html(self.kp)

            self._widget_model_summary.value = self._get_model_summary_html(self.kp)

    def _read_knowledgepack_description(self, kp, widget_tab):
        html_list = []
        knowledgepack_description = kp.knowledgepack_description
        columns = ["Category", "Generator", "Sensors"]
        df = DataFrame(columns=columns)
        for child_model in knowledgepack_description.keys():
            temp_uuid = knowledgepack_description[child_model]["uuid"]
            temp_kp = self._dsk.pipeline.get_knowledgepack(temp_uuid)

            if widget_tab == "feature_summary":
                df = DataFrame(temp_kp.feature_summary)[columns]
                df_html = df.to_html()
            elif widget_tab == "model_summary":
                df_html = self._get_model_summary(temp_kp)
            elif widget_tab == "confusion_matrix":
                df_html = temp_kp.confusion_matrix_by_metric("validation").__html__()

            report = str(
                [
                    v + ": " + temp_kp.class_map[i]
                    for (i, v) in knowledgepack_description[child_model][
                        "results"
                    ].items()
                ]
            )

            df_text = (
                "<br>"
                + '<h1 style="font-size:300%;">'
                + child_model
                + " </h1>"
                + '<h1 style="font-size:150%;"> Class Map: '
                + report
                + "</h1>"
            )

            html_list.append(df_text + df_html)

        return "".join(html_list)

    def _get_confusion_matrix_html(self, kp):
        if self.kp.knowledgepack_description:
            df_text = (
                "<br>"
                + '<h1 style="font-size:300%;"> Confusion Matrix : Feature Recall </h1>'
            )
            html_list = (
                df_text + self.kp.confusion_matrix_by_metric("hierarchical").__html__()
            )
            return html_list + self._read_knowledgepack_description(
                kp, "confusion_matrix"
            )
        else:
            return self.kp.confusion_matrix_by_metric("validation").__html__()

    def _get_feature_summary_html(self, kp):
        if kp.knowledgepack_description:
            return self._read_knowledgepack_description(kp, "feature_summary")
        else:
            return DataFrame(kp.feature_summary)[
                ["Category", "Generator", "Sensors"]
            ].to_html()

    def _get_model_summary_html(self, kp):
        if kp.knowledgepack_description:
            return self._read_knowledgepack_description(kp, "model_summary")
        else:
            return self._get_model_summary(kp)

    def _get_model_summary(self, kp):
        model_summary = kp.device_configuration
        model_summary.pop("sample_rate", None)
        model_summary["uuid"] = kp.uuid

        columns = list(model_summary.keys())

        if kp.device_configuration["classifier"] == "PME":
            important_metrics = [
                "classifier",
                "neurons",
                "classification_mode",
                "distance_mode",
                "optimizer",
            ]
            model_summary["optimizer"] = kp.pipeline_summary[-1]["optimizers"][0][
                "name"
            ]
            model_summary["neurons"] = len(kp.neuron_array)
            model_summary["distance_mode"] = (
                "LSUP" if model_summary["distance_mode"] else "L1"
            )
            model_summary["classification_mode"] = (
                "KNN" if model_summary["classification_mode"] else "RBF"
            )
            columns = [x for x in model_summary.keys() if x not in important_metrics]
            columns = important_metrics + columns

        if kp.device_configuration["classifier"] == "Decision Tree Ensemble":
            model_summary["n_estimators"] = kp.pipeline_summary[-1]["optimizers"][0][
                "inputs"
            ]["n_estimators"]
            model_summary["max_depth"] = kp.pipeline_summary[-1]["optimizers"][0][
                "inputs"
            ]["max_depth"]
            columns = ["classifier", "n_estimators", "max_depth", "uuid"]

        if kp.device_configuration["classifier"] == "Boosted Tree Ensemble":
            # TODO: Parse boosted tree ensemble
            columns = list(model_summary.keys())
            pass

        model_summary_df = DataFrame([model_summary])

        return model_summary_df[columns].to_html()

    def _refresh_models_list(self, b):
        if self._dsk:
            if self._dsk.pipeline:
                self._widget_parent_select.options = self.get_kp_dict()
                self._widget_parent_select.value = self._widget_parent_select.options[
                    0
                ][1]

    def create_widget(self):

        description_style = {"description_width": "112px"}

        self._info_validation_method_size = widgets.Button(
            icon="question",
            disabled=True,
            tooltip="Uses a majority vote algorithm over past events in the buffer to determine the current classification. A buffer size of 1 is equivalent to no post processing.",
            layout=Layout(width="5%", align_self="flex-end"),
        )
        
        self._button_recognize_signal_capturefile = Button(
            icon="play",
            tooltip="Run recognition against the provided test data by using selected model.",
            layout=Layout(width="20%"),
            description="Classify",
        )

        self._button_clear_the_cache = Button(
            icon="remove",
            tooltip="Delete the cache data for the selected file.",
            layout=Layout(width="5%"),
        )

        self._button_refresh = Button(
            icon="refresh", layout=Layout(width="15%"), tooltip="Refresh Model List"
        )

        self._check_show_features = Checkbox(description="Plot Features", value=False)
        self._check_show_feature_values = Checkbox(
            description="Feature Values", value=False
        )

        self._button_rehydrate = Button(
            icon="copy",
            layout=Layout(width="15%"),
            tooltip='Rehydate and paste pipeline coded. Note ( The copied code will have extra """ wrapping the code, remove those before executing',
        )
        self._widget_parent_select = Dropdown(
            description="Model Name", options=[], layout=Layout(width="85%")
        )
        self._widget_test_data = Dropdown(description="Test Data", options=[None])
        self._widget_confusion_matrix = HTML(disabled=True)
        self._widget_feature_summary = HTML(disabled=True)
        self._widget_model_summary = HTML(disabled=True)
        self._widget_render_space = HTML(disabled=True)

        self._w_file_list = Dropdown(
            description="Select File Type",
            options=["captures", "datafiles"],
            value="captures",
            style={"description_width": "112px"},
        )

        self._html_capture_explanation = HTML(
            "Evaluate your model by selecting one or more capture files for classification."
        )

        self._widget_capture_data = qgrid.show_grid(
            DataFrame({"capture_name": [""]}),
            grid_options={
                "rowHeight": 32,
                "maxVisibleRows": 5,
                "minVisibleRows": 2,
                "editable": False,
                "defaultColumnWidth": 100,
                "forceFitColumns": True,
                "sortable": True,
                "filterable": True,
            },
            column_options={"minWidth": 30},
            column_definitions={"index": {"width": 30}, "capture_name": {"width": 350}},
        )

        self.renderer = WidgetAttributeRenderer(self._widget_render_space, "value")

        self._output_recognition_classification = widgets.Image(
            format="png", disabled=True
        )

        self._w_post_process_steps = BoundedIntText(
            description="Post Processing",
            min=1,
            max=100,
            value=1,
            step=5,
            layout=Layout(width="30%"),
            style={"description_width": "112px"},
        )

        self._button_recognize_signal_capturefile.style.button_color = "#90EE90"

        model_explore_tabs = Tab(
            [
                VBox(
                    [
                        HBox([self._html_capture_explanation, self._w_file_list]),
                        self._widget_capture_data,
                        HBox(
                            [
                                self._button_recognize_signal_capturefile,
                                self._button_clear_the_cache,
                                self._w_post_process_steps,
                                self._info_validation_method_size,
                                self._check_show_features,
                                self._check_show_feature_values,
            
                            ],
                            layout=Layout(width="100%"),
                        ),
                        self._output_recognition_classification,
                    ]
                ),
                VBox([self._widget_confusion_matrix]),
                VBox([self._widget_feature_summary]),
                VBox([self._widget_model_summary]),
            ]
        )
        self.kb_items = VBox(
            [
                HBox(
                    [
                        self._widget_parent_select,
                        self._button_refresh,
                        self._button_rehydrate,
                    ]
                ),
                model_explore_tabs,
                self._widget_render_space,
            ]
        )

        model_explore_tabs.set_title(0, "Test Model")
        model_explore_tabs.set_title(1, "Confusion Matrix")
        model_explore_tabs.set_title(2, "Feature Summary")
        model_explore_tabs.set_title(3, "Model Summary")

        self._output_recognition_classification.layout.visibility = "hidden"

        self._button_refresh.on_click(self._refresh_models_list)
        self._button_rehydrate.on_click(self._rehydrate_model)
        self._widget_parent_select.observe(self._update_confusion_matrix)
        self._button_recognize_signal_capturefile.on_click(self._recognize_signal_all)
        self._button_clear_the_cache.on_click(self._clear_the_cache)
        self._w_file_list.observe(self._refresh_file_list)

        self._refresh()

        return self.kb_items


def post_process(results, buffer):
    if buffer == 1:
        return results

    if buffer > results.shape[0]:
        buffer = results.shape[0]

    post_processed_results = [-1] * (buffer-1)

    for i in range(0, results.shape[0] - buffer +1):
        counts = np.bincount(results["Classification"].iloc[i : i + buffer].values)
        post_processed_results.append(np.argmax(counts))
        
    results["Classification"] = post_processed_results

    return results[results["Classification"] != -1].reset_index(drop=True)


def get_classification_plot(kp, filename, results, smoothing, ytick_labels):

    ax = results.plot(
        y="Classification",
        x="SegmentStart",
        style="--o",
        yticks=range(len(ytick_labels)),
        figsize=(16, 5),
        title="Model: "
        + kp.name
        + "  Capture: "
        + filename
        + "  Smoothing: {}".format(smoothing),
    )

    _ = ax.set_yticklabels(ytick_labels)
    _ = ax.set_xlabel("time (samples)")

    fig = ax.get_figure()

    return fig


def get_classification_and_features_plot(
    kp, filename, results, smoothing, ytick_labels, show_numbers=False
):

    title = (
        "Model: "
        + kp.name
        + "  Capture: "
        + filename
        + "  Smoothing: {}".format(smoothing)
    )

    if isinstance(results.FeatureVector.loc[0], str):
        results.FeatureVector = results.FeatureVector.apply(lambda x: literal_eval(x))

    tmp_data = np.zeros((len(results), len(results.FeatureVector.loc[0])))
    for index, features in enumerate(results.FeatureVector):
        tmp_data[index] += np.array(features)

    features = [x["Feature"] for x in kp.feature_summary]
    classifications = results.ClassificationName

    ##
    num_features = len(features)
    num_classes = len(kp.class_map) + 2
    fig_size = (16, num_classes + num_features // 3)

    print(fig_size)
    fig, axs = plt.subplots(
        2,
        1,
        figsize=fig_size,
        gridspec_kw={"height_ratios": [num_classes, num_features]},
    )

    axs[0].set_title(title)
    axs[0].plot(results.SegmentStart, results.Classification, "--o")
    axs[0].set_xlim(0, results.SegmentStart.values[-1])
    axs[0].set_yticks(range(len(ytick_labels)))
    axs[0].set_yticklabels(ytick_labels)
    axs[0].set_xlabel("time (samples)")

    axs[1].imshow(tmp_data.T, cmap="Blues", aspect="auto")

    axs[1].set_xticks(np.arange(len(classifications)))
    axs[1].set_yticks(np.arange(len(features)))

    axs[1].set_xticklabels([])
    axs[1].set_yticklabels(features)

    # Rotate the tick labels and set their alignment.
    plt.setp(axs[1].get_xticklabels(), rotation=90, ha="right", rotation_mode="anchor")

    # Loop over data dimensions and create text annotations.
    if show_numbers:
        for i in range(len(features)):
            for j in range(len(classifications)):
                text = axs[1].text(
                    j, i, int(tmp_data.T[i, j]), ha="center", va="center", color="w"
                )

    fig.tight_layout()

    return fig



from ipywidgets import Layout, Button, VBox, HBox, Dropdown, Checkbox, Label, widgets, HTML, GridspecLayout
from sensiml.widgets.base_widget import BaseWidget
import qgrid
from pandas import DataFrame


class FeatureGeneratorWidget(BaseWidget):

    def _refresh(self):
        if not self._dsk:
            raise Exception("Must have dsk.")

        if not self._dsk.project:
            raise Exception("Must have active project.")

        if not self._dsk.pipeline:
            raise Exception("Must have active pipeline.")

        fill_widgets_grid(self._dsk, self._w_grid)

    def hide(self):
        self._w_grid.layout.height = '0px'

    def show(self):
        self._w_grid.layout.height='auto'

    def _clear(self):
        for index, subtype in enumerate(self._w_subtype):
            subtype.children[0].value = False

    def get_fature_generator(self):
        subtype_list = get_active_subtypes(self._w_grid)

        feature_generators = self._dsk.pipeline.add_feature_generator([{'subtype_call': str(x)} for x in subtype_list],
                                                 function_defaults={'columns':self._dsk.pipeline.data_columns},
                                                 return_generator_set=True)._to_dict()

        feature_generators['outputs'] = ['temp.generator_set0', 'temp.features.generator_set0']

        return feature_generators

    def create_widget(self):

        self._w_grid = build_widgets_grid()

        return self._w_grid

def get_subtype_defaults(dsk, subtype):
    fgs = dsk.list_functions(functype='Feature Generator',
                             subtype=subtype, kp_functions=True).df.NAME
    params = {}
    for fg in fgs:
        for param in dsk.functions._function_list[fg].input_contract:
            if param['name'] in ['input_data', 'columns', 'group_columns']:
                continue
            params[param['name']] = param['type']

    return params


def build_widgets_grid(width=3, height=6):
    grid = GridspecLayout(height, width)
    for row in range(height):
        for col in range(width):
            grid[row, col] = widgets.Checkbox(value=False,
                                              description='{}{}'.format(row,col),
                                              layout=Layout(visibility='hidden')
                                              )

    return grid


def fill_widgets_grid(dsk, grid):

    fg_subtypes = dsk.list_functions(
        functype='Feature Generator').df.SUBTYPE.unique()

    index = 0
    for row in range(grid.n_rows):
        for col in range(grid.n_columns):
            if index >= len(fg_subtypes):
                continue
            grid[row, col].value = False
            grid[row, col].description = fg_subtypes[index]
            grid[row, col].layout.visibility = 'visible'
            index += 1

    return grid


def get_active_subtypes(grid):
    active = []
    for row in range(grid.n_rows):
        for col in range(grid.n_columns):
            if grid[row, col].value:
                active.append(grid[row, col].description)

    return active


def build_pipeline(dsk, active_subtypes):
    subtype_list = []
    for subtype in active_subtypes:
        tmp_dict = {'subtype_call': subtype}
        subtype_list.append(tmp_dict)

    dsk.pipeline.add_feature_generator(subtype_list, function_defaults={
                                       'columns': dsk.pipeline.data_columns})


def build_subtype_list(dsk, active_subtypes):
    subtype_list = []
    for subtype in active_subtypes:
        tmp_dict = {'subtype_call': subtype,
                    'params': get_subtype_defaults(dsk, subtype)}
        subtype_list.append(tmp_dict)

    return subtype_list



class BaseMethodCall(object):
    """The base class for calls to functions.
    Child classes have their own additional properties and overwrite this docstring.
    """

    def __init__(self, name="", function_type=""):
        self._name = name
        self._type = function_type

    def _public_properties(self):
        """Returns the object's public properties, i.e. those that do not start with _"""
        return (name for name in dir(self) if not name.startswith("_"))

    def _to_dict(self):
        prop_dict = {}
        prop_dict["inputs"] = {}
        for prop in self._public_properties():
            if getattr(self, prop) is not None:
                prop_dict["inputs"][prop] = getattr(self, prop)
        prop_dict["name"] = self._name
        return prop_dict



from glob import glob

try:
    import configparser as ConfigParser
except:
    import ConfigParser
import os.path


def get_config(config_info_file=None):
    """ Sample Use Case
    ============================================================================
    config_info_file="C:\\Users\\username\\project\\nbi_jv-nbijv\\notebooks\\upload_config.ini"
    data = get_config("your config file")
    files = data[0]   /* file path
    group_names = data[1]  /* ['Accelerator', 'Gyroscrop']
    group_columns = data[2]  /* ['3,4,5', '6,7,8']
    data_column_names = data[3]  /* ['accelx,accely,accelz', 'gyrox,gyroy,gyroz']
    metadata_columns = data[4]  /* [0,1]
    """
    if config_info_file is None:
        raise Exception(
            "*** Sorry, Config File Missing *** "
            + "file exist? "
            + str(os.path.isfile(config_info_file))
            + "..filename: "
            + config_info_file
        )

    config = ConfigParser.ConfigParser()
    config.read(config_info_file)
    sections = config.sections()
    if "File_Path" in sections[0]:
        filePath = config.get(sections[0], "path")
    else:
        print("Can not read config file...")

    # First try absolute path
    files = glob(filePath)
    if not files:
        # Then try relative to upload config file
        files = glob(os.path.join(os.path.split(config_info_file)[0], filePath))

    for i in range(len(files)):
        if os.path.splitext(files[i])[1] == ".h5":
            return [files, None, None, None, None]

    group_names = config.get(sections[1], "group_name")
    group_columns = config.get(sections[1], "group_column")
    data_column_names = config.get(sections[1], "data_column_name")
    metadata_columns = (
        (config.get(sections[2], "metadata_column")).split(",")
        if len(sections) > 2
        else None
    )
    groupnames = group_names.split("/")
    groupcolumns = group_columns.split("/")
    datacolumnnames = data_column_names.split("/")
    # metadatacolumns = metadata_columns.split(',')

    return [files, groupnames, groupcolumns, datacolumnnames, metadata_columns]



import json
from numpy import float64, int64
import sensiml.base.utility as utility
import logging
import time


logger = logging.getLogger(__name__)


class MetadataType(object):
    Int = "integer"
    Float = "float"
    String = "string"


class Metadata(object):
    """Base class for a metadata object."""

    def __init__(self, connection, project, capture):
        """Initialize a metadata object.

            Args:
                connection
                project
                capture
        """
        self._uuid = ""
        self._name = ""
        self._sample_start = 0
        self._sample_end = 0
        self._value = ""
        self._value_type = "string"
        self._comments = ""
        self._connection = connection
        self._project = project
        self._capture = capture
        self._lastmodifiedby = ""
        self._lastmodifiedtime = ""

    @property
    def uuid(self):
        """Auto generated unique identifier for the metadata object"""
        return self._uuid

    @uuid.setter
    def uuid(self, value):
        self._uuid = value

    @property
    def name(self):
        """The name property of the metadata object"""
        return self._name

    @name.setter
    def name(self, value):
        self._name = value

    @property
    def value(self):
        """The value of the metadata object

        Note:
            The metadata value -1 (as an integer or string) has a special meaning. It is reserved by KnowledgeBuilder
            and means that the metadata name is not defined for a particular sample. You should never use -1 for a
            metadata value.
        """
        if self._value_type == MetadataType.Float:
            return float(self._value)
        elif self._value_type == MetadataType.Int:
            return int(self._value)
        else:
            return self._value

    @value.setter
    def value(self, value):
        if type(value) in (float, float64):
            self._value_type = MetadataType.Float
            self._value = str(value)
        elif type(value) in (int, int64):
            self._value_type = MetadataType.Int
            self._value = str(value)
        else:
            self._value_type = MetadataType.String
            self._value = str(value)

    @property
    def value_type(self):
        """The data type of the metadata object"""
        return self._value_type

    @property
    def lastmodifiedby(self):
        return self._lastmodifiedby

    @property
    def lastmodifiedtime(self):
        return self._lastmodifiedtime

    @property
    def comments(self):
        """The label comments"""
        return self._comments

    @comments.setter
    def comments(self, value):
        self._comments = value

    def insert(self):
        """Calls the REST API and inserts a metadata object onto the server using the local object's properties."""
        self._capture.await_ready()
        url = "v2/project/{0}/capture/{1}/metadata-relationship/".format(
            self._project.uuid, self._capture.uuid
        )
        label_info = {
            "name": self.name,
            "type": self.value_type,
            "value": self.value,
            "comments": self._comments,
        }
        response = self._connection.request("post", url, label_info)
        response_data, err = utility.check_server_response(response)
        if err is False:
            self.uuid = response_data["uuid"]

    def update(self):
        """Calls the REST API and updates the object on the server."""
        self._capture.await_ready()
        url = "v2/project/{0}/capture/{1}/metadata-relationship/{2}/".format(
            self._project.uuid, self._capture.uuid, self.uuid
        )
        label_info = {
            "name": self.name,
            "type": self.value_type,
            "value": self.value,
            "comments": self._comments,
        }
        response = self._connection.request("put", url, label_info)
        response_data, err = utility.check_server_response(response)

    def delete(self):
        """Calls the REST API and deletes the object from the server."""
        url = "v2/project/{0}/capture/{1}/metadata-relationship/{2}/".format(
            self._project.uuid, self._capture.uuid, self.uuid
        )
        response = self._connection.request("delete", url)
        response_data, err = utility.check_server_response(response)

    def refresh(self):
        """Calls the REST API and populates the local object's properties from the server."""
        url = "v2/project/{0}/capture/{1}/metadata-relationship/{2}/".format(
            self._project.uuid, self._capture.uuid, self.uuid
        )
        response = self._connection.request("get", url)
        response_data, err = utility.check_server_response(response)
        if err is False:
            self.initialize_from_dict(response_data)

    def initialize_from_dict(self, dict):
        """Reads a json dictionary and populates a single metadata object.

            Args:
                dict (dict): contains the uuid, name, type, value, properties
        """
        self.uuid = dict["uuid"]
        self.name = dict["name"]
        if dict["type"] == MetadataType.Int:
            self.value = int(dict["value"])
        elif dict["type"] == MetadataType.Float:
            self.value = float(dict["value"])
        else:
            self.value = dict["value"]



from six import string_types
from sensiml.datamanager.pipeline import PipelineStep


class FunctionCall(PipelineStep):
    """The base class for a function call.

    Child classes have their own additional properties and overwrite this docstring."""

    def __init__(self, name, function_type="transform"):
        super(FunctionCall, self).__init__(name=name, step_type="FunctionCall")
        self._type = function_type

    @property
    def outputs(self):
        return self._outputs

    @outputs.setter
    def outputs(self, outputs):
        self._outputs = outputs

    def _check_inputs(self, output_vars=[]):
        """Compares the list of inputs with the given output_vars list

            Args:
                output_vars (list[str]): the current set of output variables

            Returns:
                unmatched_inputs (list[str]): unmatched inputs or an empty list if all inputs are matched
        """
        unmatched_inputs = []
        for param in [getattr(self, input["name"]) for input in self._input_contract]:
            if (
                isinstance(param, string_types)
                and param.find("temp.") > -1
                and not param in output_vars
            ):
                unmatched_inputs.append(param)
        return unmatched_inputs

    def _public_properties(self):
        """Returns the object's public properties, i.e. those that do not start with _"""
        return (name for name in dir(self) if not name.startswith("_"))

    def _to_dict(self):
        prop_dict = {}
        prop_dict["name"] = self._name
        prop_dict["type"] = self._type
        prop_dict["feature_table"] = self._feature_table
        prop_dict["inputs"] = {}
        for prop in [
            input["name"]
            for input in self._input_contract
            if not input.get("handle_by_set", False)
        ]:
            if prop != "outputs":
                prop_dict["inputs"][prop] = getattr(self, prop)
        prop_dict["outputs"] = getattr(self, "outputs")
        return prop_dict



from sensiml.device_flashing.device_flasher import DeviceFlasher
from sensiml.device_flashing.stlink_device_flasher import STDeviceFlasher
from sensiml.device_flashing.nordic_device_flasher import NordicThingyFlasher


__all__ = ['STDeviceFlasher', 'DeviceFlasher', 'NordicThingyFlasher']


import logging
import time
import datetime
from requests.exceptions import *
import sensiml.base.utility as utility
from sensiml.datamanager.errors import CaptureUploadFailureError
from sensiml.datamanager.metadataset import MetadataSet
from sensiml.datamanager.segment import SegmentSet
import pandas as pd
from pandas import DataFrame
import os.path

EXCEPTION_STATES = frozenset(["FAILURE", "RETRY", "REVOKED"])

# Custom ready states because sync or nonexistent task state is set to NONE
READY_STATES = frozenset(["FAILURE", "REVOKED", "SUCCESS", None])

# Custom unready states because celery's PENDING state is unreliable
UNREADY_STATES = frozenset(["STARTED", "RECEIVED", "RETRY", "SENT"])


logger = logging.getLogger(__name__)


class CaptureTaskFailedError(Exception):
    """Raised if Capture task failed"""


class Capture(object):
    """Base class for an Capture."""

    def __init__(
        self,
        connection,
        project,
        filename="",
        path="",
        uuid="",
        capture_configuration_uuid=None,
        capture_info=None,
        created_at=None,
        **kwargs
    ):
        """Initialize an Capture instance."""
        self._connection = connection
        self._project = project
        self._filename = filename
        self._created_at = created_at
        self._path = path
        self._uuid = uuid
        self._capture_info = capture_info
        self._metadata = MetadataSet(self._connection, self._project, self)
        self._segements = SegmentSet(self._connection, self._project, self, initialize_set=False)

    @property
    def uuid(self):
        """Auto generated unique identifier for the Capture object"""
        return self._uuid

    @property
    def filename(self):
        return self._filename

    @filename.setter
    def filename(self, value):
        self._filename = value

    @property
    def created_at(self):
        return self._created_at

    @created_at.setter
    def created_at(self, value):
        if value:
            self._created_at = datetime.datetime.strptime(
                value[:-6], "%Y-%m-%dT%H:%M:%S.%f"
            )

    @property
    def path(self):
        """The local or server path to the Capture file data"""
        return self._path

    @path.setter
    def path(self, value):
        self._path = value

    @property
    def metadataset(self):
        return self._metadata

    @property
    def segmentset(self):
        return self._segmentset

    @property
    def capture_info(self):
        """Info about capture"""
        return self._capture_info

    @capture_info.setter
    def capture_info(self, value):
        self._capture_info = {}
        if value.get("CalculatedSampleRate", None) is not None:
            self._capture_info["calculated_sample_rate"] = value.get(
                "CalculatedSampleRate"
            )
        if value.get("SampleRate", None) is not None:
            self._capture_info["set_sample_rate"] = value.get(
                "SampleRate", None)
        if value.get("capture_configuration_uuid", None) is not None:
            self._capture_info["capture_configuration_uuid"] = value.get(
                "capture_configuration_uuid", None
            )

        if not self._capture_info:
            self._capture_info = None

    @property
    def ready(self):
        """Returns if Capture (Capture) is ready or not

        Returns:
            Boolean: True if task is ready or False if task is pending. Raises Exception if task failed.
        """
        url = "project/{0}/capture/{1}/".format(self._project.uuid, self.uuid)
        response = self._connection.request("get", url)
        response_data, err = utility.check_server_response(response)
        if (
            not response_data["task_state"]
            or response_data["task_state"] in READY_STATES
        ):
            return True

        if response_data["task_state"] in EXCEPTION_STATES:
            logger.error(response_data["task_result"])
            raise CaptureTaskFailedError()

        return False

    def await_ready(self, sleep=3, retries=0):
        """Blocks until Capture (Capture) is ready or failed

        Args:
            sleep (int): Number of seconds to sleep between polling
            retries (int): Number of times to retry before unblocking and returning False.
                           Use 0 for infinite.

        Returns:
            None or raises CaptureUploadFailureError if upload failed.
        """
        try_ = 0
        url = "project/{0}/capture/{1}/".format(self._project.uuid, self.uuid)

        while retries is 0 or try_ <= retries:
            try_ += 1
            response = self._connection.request("get", url)
            response_data, err = utility.check_server_response(response)
            # Uses custom UNREADY_STATES because we do NOT want to sleep forever on "PENDING"
            if response_data["task_state"] in UNREADY_STATES:
                time.sleep(sleep)
            elif (
                not response_data["task_state"]
                or response_data["task_state"] in READY_STATES
            ):
                return True
            else:
                raise CaptureUploadFailureError(response_data["task_result"])

    def insert(self, asynchronous=False):
        """Calls the REST API to insert a new Capture."""
        url = "project/{0}/capture/".format(self._project.uuid)
        capture_info = {"name": self.filename, "asynchronous": asynchronous}
        if self.capture_info is not None:
            capture_info.update(self._capture_info)
        response = self._connection.file_request(
            url, self.path, capture_info, "rb")
        response_data, err = utility.check_server_response(response)
        if err is False:
            self._uuid = response_data["uuid"]

    def update(self):
        """Calls the REST API to update the capture."""
        url = "project/{0}/capture/{1}/".format(self._project.uuid, self._uuid)

        capture_info = {"name": self.filename}
        if self._capture_info is not None:
            capture_info.update(self._capture_info)
        if self.path:
            response = self._connection.file_request(
                url, self.path, capture_info, "rb", "patch"
            )
        else:
            response = self._connection.request(url, "post", capture_info)
        response_data, err = utility.check_server_response(response)

    def delete(self):
        """Calls the REST API to delete the capture from the server."""
        url = "project/{0}/capture/{1}/".format(self._project.uuid, self._uuid)
        response = self._connection.request("delete", url)
        response_data, err = utility.check_server_response(response)

    def refresh(self):
        """Calls the REST API and self populates properties from the server."""
        url = "project/{0}/capture/{1}/".format(self._project.uuid, self._uuid)
        response = self._connection.request("get", url)
        response_data, err = utility.check_server_response(response)
        if err is False:
            self.filename = response_data["name"]

    @classmethod
    def initialize_from_dict(cls, capture_dict):
        """Reads a dictionary or properties and populates a single capture.

            Args:
                capture_dict (dict): contains the capture's 'name' property

            Returns:
                capture object
        """
        assert isinstance(capture_dict, dict)
        new_dict = capture_dict.copy()
        new_dict["filename"] = str(
            new_dict.get(
                "filename",  # capture object attribute
                new_dict.pop("name", ""),  # API object key
            )
        )
        return Capture(**new_dict)

    def data(self, cloud=True):
        """Calls the REST API and retrieves an capture's sensor data.

            Returns:
                data (DataFrame): all sensor data associated with the capture
        """
        url = "project/{0}/capture/{1}/capture-sample/".format(
            self._project.uuid, self.uuid
        )
        data = []

        if cloud:
            try:
                response = self._connection.request("get", url)
                response_data, err = utility.check_server_response(response)
                if err is False:
                    data = DataFrame(response_data)
            except HTTPError:
                utility.read_response_json_data(response)
                raise

        else:
            try:
                paged_response = self._connection.paged_request("get", url)
                for results in paged_response:
                    data.append(DataFrame(results))
            except HTTPError:
                utility.read_response_json_data(paged_response.response)
                raise
            data = pd.concat(data)

        # Post-processing to get numeric index
        data["sequence"] = data.sequence.astype(int)
        data = data.sort_values(by="sequence").reset_index(drop=True)
        data.drop("sequence", axis=1, inplace=True)

        return data

    # methods to make this instance hashable and comparable
    def __hash__(self):
        return hash(self._uuid)

    def __eq__(self, other):
        return (self._uuid,) == (other._uuid,)

    def __ne__(self, other):
        return not self == other

    def __lt__(self, other):
        return (self._uuid,) < (other._uuid,)

    def __gt__(self, other):
        return (self._uuid,) > (other._uuid,)

    def __le__(self, other):
        return (self < other) or (self == other)

    def __ge__(self, other):
        return (self > other) or (self == other)



import json
from numpy import float64, int64
import sensiml.base.utility as utility
import logging
import time


logger = logging.getLogger(__name__)


class LabelType(object):
    Int = "integer"
    Float = "float"
    String = "string"


class Label(object):
    """Base class for a label object."""

    def __init__(self, connection, project):
        """Initialize a metadata object.

            Args:
                connection
                project
        """
        self._uuid = ""
        self._name = ""
        self._value_type = "string"
        self._metadata = False
        self._is_dropdown = None
        self._last_modified = ""
        self._created_at = ""
        self._connection = connection
        self._project = project

    @property
    def uuid(self):
        """Auto generated unique identifier for the metadata object"""
        return self._uuid

    @uuid.setter
    def uuid(self, value):
        self._uuid = value

    @property
    def name(self):
        """The name property of the metadata object"""
        return self._name

    @name.setter
    def name(self, value):
        self._name = value

    @property
    def value_type(self):
        """The data type of the metadata object"""
        return self._value_type

    @value_type.setter
    def value_type(self, value):
        self._value_type = value

    def insert(self):
        """Calls the REST API and inserts a metadata object onto the server using the local object's properties."""
        url = "project/{0}/label/".format(
            self._project.uuid
        )
        label_info = {
            "name": self.name,
            "type": self.value_type,
            "is_dropdown": self.is_dropdown,
            "metadata": self.metadata
        }
        response = self._connection.request("post", url, label_info)
        response_data, err = utility.check_server_response(response)
        if err is False:
            self.uuid = response_data["uuid"]

    def update(self):
        """Calls the REST API and updates the object on the server."""
        url = "project/{0}/label/{1}/".format(
            self._project.uuid, self.uuid
        )
        label_info = {
            "name": self.name,
            "type": self.value_type,
            "value": self.value,
            "is_dropdown": self.is_dropdown
        }
        response = self._connection.request("put", url, label_info)
        response_data, err = utility.check_server_response(response)

    def delete(self):
        """Calls the REST API and deletes the object from the server."""
        url = "project/{0}/label/{1}/".format(
            self._project.uuid, self.uuid
        )
        response = self._connection.request("delete", url)
        response_data, err = utility.check_server_response(response)

    def refresh(self):
        """Calls the REST API and populates the local object's properties from the server."""
        url = "project/{0}/label/{1}/".format(
            self._project.uuid, self.uuid
        )
        response = self._connection.request("get", url)
        response_data, err = utility.check_server_response(response)
        if err is False:
            self.initialize_from_dict(response_data)

    def initialize_from_dict(self, data):
        """Reads a json dictionary and populates a single metadata object.

            Args:
                data (dict): contains the uuid, name, type
        """
        self.uuid = data["uuid"]
        self.name = data["name"]
        if data["type"] == LabelType.Int:
            self.value_type = int(data["type"])
        elif data["type"] == LabelType.Float:
            self.value_type = float(data["type"])
        else:
            self.value_type = data["type"]


class LabelSet(object):

    def __init__(self, connection, project, initialize_set=True):
        """Initialize a metadata object.

            Args:
                connection
                project
        """
        self._connection = connection
        self._project = project
        self._set = None

        if initialize_set:
            self.refresh()

    @property
    def labels(self):
        if self._set is None:
            self._set = self.get_set()

        return self._set

    def refresh(self):
        self._set = self.get_set()


    def get_set(self):
        """Calls the REST API and inserts a metadata object onto the server using the local object's properties."""
        url = "project/{0}/label/".format(
            self._project.uuid
        )

        response = self._connection.request("get", url)
        response_data, err = utility.check_server_response(response)
        if err:
            raise Exception(err)

        # Populate each label from the server
        labels = []
        for obj in response_data:
            labels.append(self._new_obj_from_dict(obj))

        return labels

    def _new_obj_from_dict(self, data):
        """Creates a new label from data in the dictionary.

            Args:
                data (dict): contains label properties uuid, name, type

            Returns:
                label object

        """
        obj = Label(self._connection, self._project)
        obj.initialize_from_dict(data)
        return obj



import json
import datetime
from pandas import DataFrame
import re
import warnings
import types

import sensiml.base.utility as utility


class QueryColumns(object):
    """Base class for the query columns and metadata_columns properties"""

    def __init__(self, columns=[]):
        if len(columns) > 0:
            self._columns = columns
        else:
            self._columns = []

    def add(self, *items):
        """Adds a column name or names for query selection.

            Args:
                items (str or list[str]): column names to add

            Note:
                The named column(s) must exist in the project and duplicate column names will be ignored
        """
        for item in items:
            if item not in self._columns:
                self._columns.append(item)

    def remove(self, item):
        """Removes a column name from query selection.

            Args:
                item (str): column name to remove
        """
        if item in self._columns:
            self._columns.remove(item)

    def clear(self):
        """Clears all column names from query selection"""
        self._columns = []

    def __str__(self):
        return json.dumps(self._columns)


class Query(object):
    """Base class for a query.

    Queries extract project data, or a subset of project data, for use in a pipeline. The query must specify which
    columns of data to extract and what filter conditions to apply, and KnowledgeBuilder takes care of the database
    joins required to produce a clean, analysis-ready data frame.
    """

    def __init__(self, connection, project):
        """Initializes a query instance"""
        self._uuid = ""
        self._name = ""
        self._columns = QueryColumns()
        self._metadata_columns = QueryColumns()
        self._metadata_filter = ""
        self._segmenter = None
        self._label_column = ""
        self._combine_labels = None
        self._dirty = False
        self._connection = connection
        self._project = project
        self._created_at = None

    @property
    def uuid(self):
        return self._uuid

    @uuid.setter
    def uuid(self, value):
        self._uuid = value

    @property
    def name(self):
        """Name of the query"""
        return self._name

    @name.setter
    def name(self, value):
        self._dirty = True
        self._name = value

    @property
    def columns(self):
        """Sensor columns to include in the query result

        Note:
            Columns must correspond to actual project sensor columns or the reserved word 'SequenceID' for the
            original sample index.
        """
        self._dirty = True
        return self._columns

    @property
    def created_at(self):
        """Date of the Pipeline creation"""
        return self._created_at

    @created_at.setter
    def created_at(self, value):
        self._created_at = datetime.datetime.strptime(
            value[:19], "%Y-%m-%dT%H:%M:%S")

    @property
    def label_column(self):
        """Label columns to use in the query result

        Note:
            Columns must correspond to actual project label column"""

        self._dirty = True
        return self._label_column

    @label_column.setter
    def label_column(self, value):
        self._dirty = True
        self._label_column = value

    @property
    def combine_labels(self):
        """Combine label values into new value to use in the query result

        Label = Gesture
        Label_Values = A,B,C,D,E
        combine_labels = {'Group1':['A','B',C'],'Group2':['D','E']}

        the labels that will be returned will be group1 and group2
        """

        self._dirty = True
        return self._combine_labels

    @combine_labels.setter
    def combine_labels(self, value):
        self._dirty = True
        self._combine_labels = value

    @property
    def metadata_columns(self):
        """Metadata columns to include in the query result

        Note:
            Columns must correspond to actual project metadata columns.
        """
        self._dirty = True
        return self._metadata_columns

    @property
    def metadata_filter(self):
        """Filter criteria of the query

        Args:
            value (str): similar to a SQL WHERE clause, the string can contain any number of AND-concatenated
              expressions where square brackets surround the column name and comparison value, with the operator in
              between. Supported operators: >, >=, <, <=, =, !=, IN

        Examples::

            metadata_filter = '[Subject] > [5] AND [Subject] <= [15]'
            metadata_filter = '[Gender] = [Female] AND [Activity] != [Walking]'
            metadata_filter = '[Subject] IN [5, 7, 9, 11, 13, 15]'

        Note:
            Queries do not support OR-concatenation between expressions, but often the IN operator can be used to
            achieve OR-like functionality on a single column. For example::

                [Gesture] IN [A, M, L]

            is equivalent to::

                [Gesture] = [A] OR [Gesture] = [M] OR [Gesture] = [L]

        """
        return self._metadata_filter

    @metadata_filter.setter
    def metadata_filter(self, value):
        self._dirty = True
        self._metadata_filter = value

    @property
    def segmenter(self):
        """Segmenter to use for the query

        Args:
            value (int): ID of segmenter.

        """
        return self._segmenter

    @segmenter.setter
    def segmenter(self, value):
        if value is None or isinstance(value, int):
            self._segmenter = value
        else:
            raise ValueError("Value must be a integer")

    def insert(self, renderer=None):
        """Calls the REST API and inserts a new query."""
        url = "project/{0}/query/".format(self._project.uuid)

        query_info = {
            "name": self.name,
            "columns": self.columns._columns,
            "metadata_columns": self.metadata_columns._columns,
            "metadata_filter": self.metadata_filter,
            "segmenter_id": self._segmenter,
            "label_column": self.label_column,
            "combine_labels": self.combine_labels,
        }

        request = self._connection.request("post", url, query_info)
        response, err = utility.check_server_response(
            request, renderer=renderer)
        if err is False:
            self.uuid = response["uuid"]
            self._dirty = False

    def update(self, renderer=None):
        """Calls the REST API and updates the query object on the server."""
        url = "project/{0}/query/{1}/".format(self._project.uuid, self.uuid)

        self._checked = None
        query_info = {
            "name": self.name,
            "columns": self.columns._columns,
            "metadata_columns": self.metadata_columns._columns,
            "metadata_filter": self.metadata_filter,
            "segmenter_id": self._segmenter,
            "label_column": self.label_column,
            "combine_labels": self.combine_labels,
        }

        if len(query_info["columns"]) > 0 or len(query_info["metadata_columns"]) > 0:
            filters = (
                self._metadata_filter.split("AND")
                if len(self._metadata_filter) > 0
                else []
            )
            for i, filter in enumerate([filter for filter in filters if len(filters)]):
                match = re.search(
                    "\[(?P<key>[0-9A-Za-z_\-\. \)\(]+)\]([ ]*(?P<symbol>[\>\=|\<\=|\=|\<\>|\>|\<|in|IN|!=]+)[ ]*\[(?P<value>.+)\])?",
                    str(filter),
                )
                if not match:
                    self._checked = 1
            if self._checked is None:
                request = self._connection.request("put", url, query_info)
                response, err = utility.check_server_response(
                    request, renderer=renderer
                )
            else:
                self._metadata_filter = ""
                print("Metadata Filter is not formatted correctly!")
        else:
            print(
                "Sensor columns and, or metadata columns must be specified in order to query!"
            )

    def delete(self, renderer=None):
        """Calls the REST API and deletes the query object from the server."""
        url = "project/{0}/query/{1}/".format(self._project.uuid, self.uuid)
        request = self._connection.request("delete", url)
        response, err = utility.check_server_response(
            request, renderer=renderer)
        self._dirty = False
        if err is False:
            print(response["result"])

    def statistics_segments(self, renderer=None):
        """Returns metadata statistics for the query."""

        url = "project/{0}/query/{1}/statistics/".format(
            self._project.uuid, self.uuid)
        request = self._connection.request("get", url)
        response_data, err = utility.check_server_response(
            request, renderer=renderer)
        if err is False:
            return DataFrame(response_data)
        else:
            return None

    def statistics(self, renderer=None):
        """Returns metadata statistics for the query."""
        if self._dirty:
            self.update(renderer=renderer)

        url = "project/{0}/query/{1}/statistics/".format(
            self._project.uuid, self.uuid)
        request = self._connection.request("get", url)
        response_data, err = utility.check_server_response(
            request, renderer=renderer)
        if err is False:
            df = DataFrame(response_data)
            df.drop("Segment Length", axis=1, inplace=True)
            data = (
                df.groupby(
                    ["Capture"] + self.metadata_columns._columns +
                    [self.label_column]
                )[self.label_column]
                .size()
                .unstack(fill_value=0)
            )

            return data
        else:
            return None

    def data(self):
        """Calls the REST API for query execution and returns the result.

        Note:
            Intended for previewing the query result before creating a query call object and using it in a sandbox
            step. The resulting DataFrame is not cached on the server, but once it is used in a sandbox it may be
            cached.
        """
        warnings.warn("This call has a timeout of two minutes.")

        if self._dirty:
            self.update()

        url = "project/{0}/query/{1}/data/".format(
            self._project.uuid, self.uuid)
        request = self._connection.request("get", url)
        response_data, err = utility.check_server_response(request)
        if err is False:
            error_report = None
            if not isinstance(response_data, list):
                data = DataFrame(response_data)
            else:
                try:
                    data = DataFrame(response_data[0])
                    error_report = response_data[1]
                except:
                    data = DataFrame(response_data)

            # Post-processing to get numeric index
            data["sample"] = data.index.astype(int)
            data = data.set_index("sample")
            data.index.names = [None]
            data = data.sort_index(axis=0)
            cols = data.columns.tolist()
            if "id" in cols:
                data = data[["id"] + [c for c in cols if c != "id"]]

            return data.sort_index(axis=0), error_report
        else:
            return None

    def size(self):
        """Returns the size of the dataframe which would result from the query."""
        # Sync the client-side query with server-side query object before we check the size
        if self._dirty:
            self.update()

        url = "project/{0}/query/{1}/size/".format(
            self._project.uuid, self.uuid)
        request = self._connection.request("get", url)
        response_data, err = utility.check_server_response(request)
        if not err:
            query_result = response_data
        else:
            query_result = None
        return query_result

    def refresh(self):
        """Calls the REST API and self populate using the uuid."""
        url = "project/{0}/query/{1}/".format(self._project.uuid, self.uuid)
        request = self._connection.request("get", url)
        response, err = utility.check_server_response(request)
        if err is False:
            self.uuid = response["uuid"]
            self.name = response["name"]
            self._columns = QueryColumns(response["columns"])
            self._label_column = response["label_column"]
            self._metadata_columns = QueryColumns(response["metadata_columns"])
            self.metadata_filter = response["metadata_filter"]
            self.segmenter = response["segmenter_id"]
            self.combine_labels = response["combine_labels"]
            self._dirty = False

    def initialize_from_dict(self, query):
        """Reads a json dict and populates a single query."""
        self.uuid = query["uuid"]
        self.name = query["name"]
        self.created_at = query["created_at"]
        self._columns = QueryColumns(query["columns"])
        self._metadata_columns = QueryColumns(query["metadata_columns"])
        self.metadata_filter = query["metadata_filter"]
        self.label_column = query.get("label_column", "")
        self.segmenter = query.get("segmenter_id", None)
        self.combine_labels = query.get("combine_labels", None)

        self._dirty = False

    def post_feature_statistics(self, window_size=None):
        """Returns metadata statistics for the query."""
        data = {'window_size': window_size}
        url = "project/{0}/query/{1}/featurestatistics/".format(
            self._project.uuid, self._uuid)
        response = self._connection.request("post", url, data)

        response_data, err = utility.check_server_response(response)

        if err is False:
            return response_data

    def get_feature_statistics(self):
        """Returns metadata statistics for the query."""
        url = "project/{0}/query/{1}/featurestatistics/".format(
            self._project.uuid, self._uuid)
        response = self._connection.request("get", url)
        response_data, err = utility.check_server_response(response)

        if err is False:
            return response_data

    def __str__(self):
        return (
            "NAME: {0}\n"
            "COLUMNS: {1}\n"
            "METADATA_COLUMNS: {2}\n"
            "METADATA FILTER: {3}\n"
            "SEGMENTER ID: {4}\n"
            "LABEL COLUMN {5}\n"
            "COMBINE LABELS {6}\n"
        ).format(
            self.name,
            self.columns,
            self.metadata_columns,
            self.metadata_filter,
            self.segmenter,
            self.label_column,
            self.combine_labels,
        )

    def _to_dict(self):
        return {
            "name": self.name,
            "sensor_columns": self.columns._columns,
            "metadata_columns": self.metadata_columns._columns,
            "metadata_filter": self.metadata_filter,
            "segmenter_id": self.segmenter,
            "label_column": self.label_column,
        }



from sensiml.method_calls.generatorcall import GeneratorCall
from sensiml.datamanager.pipeline import PipelineStep


class GeneratorCallSet(PipelineStep):
    """The base class for a collection of generator calls"""

    def __init__(self, name=""):
        super(GeneratorCallSet, self).__init__(name=name, step_type="GeneratorCallSet")
        self._input_data = ""
        self._generator_calls = []
        self._group_columns = None

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        self._name = value

    @property
    def input_data(self):
        return self._input_data

    @input_data.setter
    def input_data(self, value):
        self._input_data = value

    @property
    def outputs(self):
        return self._outputs

    @outputs.setter
    def outputs(self, value):
        self._outputs = value

    @property
    def generator_calls(self):
        return self._generator_calls

    def add_generator_call(self, *generator_call):
        """Adds one or more generator calls to the collection.

            Args:
                generator_call (GeneratorCall or list[GeneratorCall]): object(s) to append
        """
        for g in generator_call:
            self._generator_calls.append(g)

    def remove_generator_call(self, *generator_calls):
        """Removes a generator call from the collection.

            Args:
                generator_calls (GeneratorCall): object to remove
        """
        for generator_call in generator_calls:
            self._generator_calls = [
                f for f in self._generator_calls if f != generator_call
            ]
        # Note:  We should probably re-populate inputs and outputs here

    @property
    def group_columns(self):
        return self._group_columns

    @group_columns.setter
    def group_columns(self, value):
        self._group_columns = value

    def _to_list(self):
        gencalls = []
        for item in self._generator_calls:
            gencalls.append(item._to_dict())
        return gencalls

    def _to_dict(self):
        gencalls_set = []
        set_dict = {}
        for item in self._generator_calls:
            gencalls_set.append(item._to_dict())
        set_dict["name"] = self._name
        set_dict["type"] = "generatorset"
        set_dict["set"] = gencalls_set
        set_dict["inputs"] = {}
        set_dict["inputs"]["group_columns"] = self._group_columns
        set_dict["inputs"]["input_data"] = self._input_data
        set_dict["outputs"] = self._outputs
        return set_dict



import os
import datetime
import json
import sensiml.base.utility as utility


class FeatureFile(object):
    """Base class for a featurefile object."""

    def __init__(self, connection, project, name="", path=""):
        self._connection = connection
        self._project = project
        self.uuid = None
        self._created_at = None
        self.name = name
        self.path = path

    # Maintain campatibility with old filename attr
    @property
    def filename(self):
        """The name of the file as stored on the server

        Note:
            Filename must contain a .csv or .arff extension
        """
        return self.name

    @filename.setter
    def filename(self, value):
        self.name = value

    @property
    def created_at(self):
        """Date of the Pipeline creation"""
        return self._created_at

    @created_at.setter
    def created_at(self, value):
        self._created_at = datetime.datetime.strptime(
            value[:-6], "%Y-%m-%dT%H:%M:%S.%f"
        )

    def insert(self):
        """Calls the REST API to insert a new featurefile."""
        url = "project/{0}/featurefile/".format(self._project.uuid)
        featurefile_info = {"name": self.name}

        if not os.path.exists(self.path):
            raise OSError(
                "Cannot update featurefile because the system cannot find the file path: '{0}'. "
                "Please update the file's path attribute.".format(self.path)
            )

        response = self._connection.file_request(url, self.path, featurefile_info, "rb")
        response_data, err = utility.check_server_response(response)
        if err is False:
            self.uuid = response_data["uuid"]

    def update(self):
        """Calls the REST API to update the featurefile's properties on the server."""
        url = "project/{0}/featurefile/{1}/".format(self._project.uuid, self.uuid)
        featurefile_info = {"name": self.name}

        if not os.path.exists(self.path):
            raise OSError(
                "Cannot update featurefile because the system cannot find the file path: '{0}'. "
                "Please update the file's path attribute.".format(self.path)
            )

        response = self._connection.file_request(
            url, self.path, featurefile_info, "rb", method="put"
        )
        response_data, err = utility.check_server_response(response)

    def delete(self):
        """Calls the REST API and deletes the featurefile from the server."""
        url = "project/{0}/featurefile/{1}/".format(self._project.uuid, self.uuid)
        response = self._connection.request("delete", url)
        response_data, err = utility.check_server_response(response)

    def refresh(self):
        """Calls the REST API and populate the featurefile's properties from the server."""
        url = "project/{0}/featurefile/{1}/".format(self._project.uuid, self.uuid)
        response = self._connection.request("get", url)
        response_data, err = utility.check_server_response(response)
        if err is False:
            self.name = response_data["name"]

    def data(self):
        """Calls the REST API and retrieves the featurefile's binary data.

        Returns:
            featurefile contents
        """
        url = "project/{0}/featurefile/{1}/data/".format(self._project.uuid, self.uuid)
        response = self._connection.request("get", url)
        return response.content

    def initialize_from_dict(self, dict):
        """Makes a new featurefile using the dictionary.

        Args:
            dict (dict): contains the featurefile's 'name' and 'uuid' properties
        """
        self.uuid = dict["uuid"]
        self.name = dict["name"]
        self.created_at = dict["created_at"]




from sensiml.datamanager.functions import Functions
from sensiml.datamanager.platformdescriptions import PlatformDescriptions
from sensiml.datamanager.pipelineseeds import PipelineSeeds
from sensiml.datamanager.featurefiles import FeatureFiles
from sensiml.datamanager.projects import Projects
from sensiml.datamanager.segmenter import SegmenterSet
from sensiml.datamanager.capture_configurations import CaptureConfigurations


__all__ = ['Functions', 'FeatureFiles', 'Projects', 'SegmenterSet',
           'PipelineSeeds', 'PlatformDescriptions', 'CaptureConfigurations']



import json
from numpy import float64, int64
import sensiml.base.utility as utility
import logging
import time


logger = logging.getLogger(__name__)


class LabelValue(object):
    Int = "integer"
    Float = "float"
    String = "string"


class LabelValue(object):
    """Base class for a label object."""

    def __init__(self, connection, project, label):
        """Initialize a metadata object.

            Args:
                connection
                project
                label
        """
        self._uuid = ""
        self._value = ""
        self._last_modified = ""
        self._created_at = ""
        self._connection = connection
        self._project = project
        self._label = label

    @property
    def uuid(self):
        """Auto generated unique identifier for the  label value object"""
        return self._uuid

    @uuid.setter
    def uuid(self, value):
        self._uuid = value

    @property
    def value(self):
        """The data type of the label value object"""
        return self._value

    @value.setter
    def value(self, value):
        """The data type of the label value object"""
        self._value = value

    @property
    def created_at(self):
        """The creatd time of the label value object"""
        return self._created_at

    def insert(self):
        """Calls the REST API and inserts a metadata object onto the server using the local object's properties."""
        url = "project/{0}/label/{1}/labelvalue/".format(
            self._project.uuid,
            self._label.uuid
        )
        data = {
            "value": self.value,
        }
        response = self._connection.request("post", url, data)
        response_data, err = utility.check_server_response(response)
        if err is False:
            self.uuid = response_data["uuid"]

    def update(self):
        """Calls the REST API and updates the object on the server."""
        url = "project/{0}/label/{1}/labelvalue/{2}".format(
            self._project.uuid,
            self._label.uuid,
            self.uuid
        )

        data = {
            "value": self.value,
        }

        response = self._connection.request("put", url, data)
        response_data, err = utility.check_server_response(response)

    def delete(self):
        """Calls the REST API and deletes the object from the server."""
        url = "project/{0}/label/{1}/labelvalue/{2}".format(
            self._project.uuid,
            self._label.uuid,
            self.uuid
        )
        response = self._connection.request("delete", url)
        response_data, err = utility.check_server_response(response)

    def refresh(self):
        """Calls the REST API and populates the local object's properties from the server."""
        url = "project/{0}/label/{1}/labelvalue/{2}".format(
            self._project.uuid,
            self._label.uuid,
            self.uuid
        )
        response = self._connection.request("get", url)
        response_data, err = utility.check_server_response(response)
        if err is False:
            self.initialize_from_dict(response_data)

    def initialize_from_dict(self, data):
        """Reads a json dictionary and populates a single metadata object.

            Args:
                dict (dict): contains the uuid, value
        """
        self.uuid = data["uuid"]
        self.value = data["value"]


class LabelValueSet(object):

    def __init__(self, connection, project, label, initialize_set=True):
        """Initialize a metadata object.

            Args:
                connection
                project
        """
        self._connection = connection
        self._project = project
        self._label = label
        self._set = None

        if initialize_set:
            self.refresh()

    @property
    def label_values(self):
        if self._set is None:
            self._set = self.get_set()

        return self._set

    def refresh(self):
        self._set = self.get_set()

    def get_set(self):
        """Calls the REST API and inserts a metadata object onto the server using the local object's properties."""
        url = "project/{0}/label/{1}/labelvalue/".format(
            self._project.uuid,
            self._label.uuid
        )

        response = self._connection.request("get", url)
        response_data, err = utility.check_server_response(response)
        if err:
            print(err)
            return None

        # Populate each label from the server
        labelvalues = []
        for obj in response_data:
            labelvalues.append(self._new_obj_from_dict(obj))

        return labelvalues

    def _new_obj_from_dict(self, data):
        """Creates a new label from data in the dictionary.

            Args:
                data (dict): contains label_value properties value, uuid

            Returns:
                label object

        """
        labelvalue = LabelValue(self._connection, self._project, self._label)
        labelvalue.initialize_from_dict(data)
        return labelvalue

    def __str__(self):
        s = ""
        if self._set:
            label = self._set[0]._label
            s = "LABEL\n"
            s += "\tname: " + str(label.name) + ' uuid: ' + \
                str(label.uuid) + "\n"
            s += "LABEL VALUES\n"
            for lbv in self._set:
                s += "\tvalue: " + str(lbv.value) + \
                    " uuid:" + str(lbv.uuid) + "\n"

        return s
